from pathlib import Path

from bx_django_utils.test_utils.fixtures import BaseFixtures, fixtures_registry


EXAMPLE_FIXTURES_DATA1 = {'info': 'This is just a silly example...', 'foo': [1, 2, 3]}
EXAMPLE_FIXTURES_DATA2 = ['Something Different', 'Test Fixtures']


BASE_PATH = Path(__file__).parent


@fixtures_registry.register()
class ExampleFixtures1(BaseFixtures):
    base_path = BASE_PATH
    file_name = 'example_fixtures1.json'

    def renew(self):
        # e.g: fetch real data from a API...
        self.store_fixture_data(data=EXAMPLE_FIXTURES_DATA1)


@fixtures_registry.register()
class ExampleFixtures2(BaseFixtures):
    base_path = BASE_PATH
    file_name = 'example_fixtures2.json'

    def renew(self):
        # e.g: fetch real data from a API...
        self.store_fixture_data(data=EXAMPLE_FIXTURES_DATA2)



import tempfile
from pathlib import Path
from typing import Union
from unittest.mock import patch

from bx_py_utils.test_utils.redirect import RedirectOut
from bx_py_utils.test_utils.snapshot import assert_text_snapshot
from django.core.management import call_command
from django.test import SimpleTestCase, TestCase

from bx_django_utils.test_utils import fixtures
from bx_django_utils.test_utils.fixtures import (
    FIXTURES_FILE_PATHS,
    BaseFixtures,
    FixturesRegistry,
    SerializerFixtures,
    fixtures_registry,
)
from bx_django_utils_tests.test_app.management.commands import renew_fixtures
from bx_django_utils_tests.test_app.models import ColorFieldTestModel
from bx_django_utils_tests.test_app.tests.fixtures.fixtures1 import (
    EXAMPLE_FIXTURES_DATA1,
    EXAMPLE_FIXTURES_DATA2,
    ExampleFixtures1,
    ExampleFixtures2,
)


class FixturesHelperTestCase(SimpleTestCase):
    def reset(self):
        FIXTURES_FILE_PATHS.clear()

    def setUp(self):
        super().setUp()
        self.reset()

    def tearDown(self):
        super().tearDown()
        self.reset()

    def test_example_fixtures(self):
        assert ExampleFixtures1().get_fixture_data() == EXAMPLE_FIXTURES_DATA1
        assert ExampleFixtures2().get_fixture_data() == EXAMPLE_FIXTURES_DATA2

    def test_registry(self):
        assert sorted(fixtures_registry.fixtures.keys()) == ['ExampleFixtures1', 'ExampleFixtures2']

    def test_str_repr(self):
        self.assertEqual(
            repr(fixtures_registry),
            '<FixturesRegistry:ExampleFixtures1,ExampleFixtures2>',
        )
        self.assertEqual(
            str(fixtures_registry),
            '<FixturesRegistry:ExampleFixtures1,ExampleFixtures2>',
        )

        fixtures = next(iter(fixtures_registry))
        self.assertIsInstance(fixtures, BaseFixtures)
        self.assertEqual(repr(fixtures), '<Fixture:ExampleFixtures1>')
        self.assertEqual(str(fixtures), '<Fixture:ExampleFixtures1>')

    def test_renew(self):
        fixture = ExampleFixtures1()

        # Change fixtures file content:
        fixture.store_fixture_data(data=['something else!'])
        assert fixture.get_fixture_data() == ['something else!']

        # "renew" -> recreate json dump back:
        fixture.renew()

        # The "old" data?
        assert fixture.get_fixture_data() == EXAMPLE_FIXTURES_DATA1

    def test_overwriting(self):
        class Foo(BaseFixtures):
            base_path = Path(__file__).parent
            file_name = 'bar.json'

        class Bar(Foo):
            pass

        registry = FixturesRegistry()
        registry.register()(Foo)

        # It's not possible to register more than two times:
        with self.assertRaisesMessage(RuntimeError, 'Fixture class "Foo" already registered!'):
            registry.register()(Foo)

        foo = registry.fixtures['Foo']
        file_path = foo.file_path

        # It's not possible to overwrite the same JSON file:
        with self.assertRaisesMessage(
            RuntimeError, f'File path "{file_path}" from "Bar" already used!'
        ):
            registry.register()(Bar)


class SerializerFixturesTestCase(TestCase):
    def test_basic(self):
        with tempfile.NamedTemporaryFile(prefix='serializer_fixtures_', suffix='.json') as tmp_file:
            tmp_file_path = Path(tmp_file.name)

            class Example(SerializerFixtures):
                """Just a test example implementation"""

                base_path = tmp_file_path.parent
                file_name = tmp_file_path.name

                def renew(self):
                    ColorFieldTestModel.objects.create(pk=1, required_color='#00AAff')
                    qs = ColorFieldTestModel.objects.all()
                    self.store_fixture_data(queryset=qs)

            example = Example()

            # Store the model instance into the JSON file:
            example.renew()

            json_str = tmp_file_path.read_text()
            self.assertIn('#00AAff', json_str)

            ColorFieldTestModel.objects.all().delete()
            self.assertEqual(ColorFieldTestModel.objects.count(), 0)

            # Maybe useful: It's possible to get the serializer data:
            data = example.get_fixture_data()
            self.assertEqual(
                data,
                [
                    {
                        'fields': {'optional_color': None, 'required_color': '#00AAff'},
                        'model': 'test_app.colorfieldtestmodel',
                        'pk': 1,
                    }
                ],
            )

            # Recreate the objects:
            instances = example.create_objects()
            self.assertEqual(ColorFieldTestModel.objects.count(), 1)
            created_instance = ColorFieldTestModel.objects.first()
            self.assertEqual(instances, [created_instance])


class FixtureMock(BaseFixtures):
    def __init__(self, name):
        self.name = name
        self.file_name = f'{name}.json'
        self.file_path = Path(f'/path/to/{name}.json')

    def renew(self):
        print(f'Mocked renew "{self.name}"')


class RegistryMock(FixturesRegistry):
    def __init__(self):
        self.fixtures = {
            'fixture1': FixtureMock('fixture1'),
            'fixture2': FixtureMock('fixture2'),
            'fixture3': FixtureMock('fixture3'),
        }


class InputMock:
    def __init__(self, input):
        self.input = input

    def __call__(self, text):
        print(text, self.input)
        return self.input


class RenewAllFixturesBaseCommandTestCase(SimpleTestCase):
    def call_command(
        self,
        *,
        options=None,
        expected_stdout: Union[str, list[str]],
        expected_stderr='',
        strip_output=True,
        input=None,
    ):
        if options is None:
            options = {}
        reg_mock = RegistryMock()
        input_mock = InputMock(input)
        with patch.object(fixtures, 'fixtures_registry', reg_mock), patch.object(fixtures, 'input', input_mock):
            with RedirectOut(strip=strip_output) as buffer:
                call_command(
                    renew_fixtures.Command(),
                    stdout=buffer._stdout_buffer,
                    stderr=buffer._stderr_buffer,
                    **options,
                )
            stderr_output = buffer.stderr
            stdout_output = buffer.stdout
            try:
                self.assertEqual(stderr_output, expected_stderr)

                if isinstance(expected_stdout, list):
                    for part in expected_stdout:
                        self.assertIn(part, stdout_output)
                else:
                    self.assertEqual(stdout_output, expected_stdout)
            except AssertionError:
                print('=' * 100)
                print('stdout:')
                print('-' * 100)
                print(stdout_output)
                print('-' * 100)
                print('stderr:')
                print('-' * 100)
                print(stderr_output)
                print('=' * 100)
                raise
        return stdout_output

    def test_renew_all(self):
        stdout_output = self.call_command(
            options=dict(all=True),
            expected_stdout=[
                '1. renew "fixture1" file "/path/to/fixture1.json ...',
                'Mocked renew "fixture1"',
                '2. renew "fixture2" file "/path/to/fixture2.json ...',
                'Mocked renew "fixture2"',
                '3. renew "fixture3" file "/path/to/fixture3.json ...',
                'Mocked renew "fixture3"',
                '3 Fixtures updated, ok.',
            ],
        )
        assert_text_snapshot(got=stdout_output)

    def test_renew_all_with_regex(self):
        stdout_output = self.call_command(
            options=dict(all=True, filter='fixture2'),
            expected_stdout=[
                '1. renew "fixture2" file "/path/to/fixture2.json ...',
                'Mocked renew "fixture2"',
                '1 Fixtures updated, ok.',
            ],
        )
        assert_text_snapshot(got=stdout_output)

    def test_select_fixtures(self):
        # Select only a few:
        stdout_output = self.call_command(
            expected_stdout=[
                '0 - fixture1 - fixture1.json',
                '1 - fixture2 - fixture2.json',
                '2 - fixture3 - fixture3.json',
                'Input one or more numbers seperated with spaces: 1 2',
                'You selection: fixture2, fixture3',
                '1. renew "fixture2" file "/path/to/fixture2.json ...',
                'Mocked renew "fixture2"',
                '2. renew "fixture3" file "/path/to/fixture3.json ...',
                'Mocked renew "fixture3"',
                '2 Fixtures updated, ok.',
            ],
            input='1 2',
        )
        assert_text_snapshot(got=stdout_output)

        # Select all by hit ENTER:
        stdout_output = self.call_command(
            expected_stdout=[
                'Renew all fixtures:',
                '3 Fixtures updated, ok.',
            ],
            input='',
        )
        assert_text_snapshot(got=stdout_output)



from django import template


register = template.Library()


@register.filter
def dict_get(value, member_name):
    """
    Returns the wanted member of a dict-like container, or an empty string
    if the member cannot be found.
    When you know the key in advance, you can access the associated value with
    DTL's regular dot-notation. `dict_get` allows you to access the value from
    a variable containing the key.
    You can also use it as a shortcut for `{{ mydict.member|default_if_none:'' }}`.

    e.g.:
        context = {'data': {'foo': '123', 'bar': '456'}}

        {% with member='bar' %}
            <span>{{ data|dict_get:member }}</span>
        {% endwith %}

        renders: <span>456</span>
    """
    return value.get(member_name, '')



"""
    Django settings for bx_django_utils test project.
"""
import os
from pathlib import Path


BASE_DIR = Path(__file__).parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'Only a test project!'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    #
    'debug_toolbar',
    #
    # Own Apps:
    'bx_django_utils',
    'bx_django_utils.approve_workflow',
    'bx_django_utils_tests.test_app',
    'bx_django_utils_tests.approve_workflow_test_app',
    #
    # Admin extra views demo:
    'bx_django_utils.admin_extra_views.apps.AdminExtraViewsAppConfig',
    'bx_django_utils.admin_extra_views.admin_config.CustomAdminConfig',
    #
    # User Timezone:
    'bx_django_utils.user_timezone.apps.UserTimezoneAppConfig',
    #
    # Feature Flags:
    'bx_django_utils.feature_flags.apps.FeatureFlagsAppConfig',
]

MIDDLEWARE = [
    'debug_toolbar.middleware.DebugToolbarMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    #
    # User Timezone:
    'bx_django_utils.user_timezone.middleware.UserTimezoneMiddleware',
]

ROOT_URLCONF = 'bx_django_utils_tests.test_project.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [str(BASE_DIR / 'test_project' / 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'bx_django_utils_tests.test_project.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': str(BASE_DIR / 'db.sqlite3'),
    }
}
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

# Password validation
AUTH_PASSWORD_VALIDATORS = []  # Just a test project, so no restrictions


# Internationalization

LANGUAGE_CODE = 'en-us'
USE_I18N = True
USE_L10N = True
LOCALE_PATHS = (
    BASE_DIR.parent / 'bx_django_utils' / 'locale',
)

TIME_ZONE = 'UTC'
USE_TZ = True
VISIBLE_TIMEZONES = ['Europe/Berlin', 'America/Los_Angeles']


# Static files (CSS, JavaScript, Images)
STATIC_URL = '/static/'
STATIC_ROOT = BASE_DIR / 'static'


INTERNAL_IPS = [
    '127.0.0.1',
]

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '{asctime} {levelname} {name} {module}.{funcName} {message}',
            'style': '{',
        },
    },
    'handlers': {'console': {'class': 'logging.StreamHandler', 'formatter': 'verbose'}},
    'loggers': {
        '': {'handlers': ['console'], 'level': 'WARNING', 'propagate': False},
        'django': {'handlers': ['console'], 'level': 'INFO', 'propagate': False},
        'django.auth': {'handlers': ['console'], 'level': 'DEBUG', 'propagate': False},
        'django.security': {'handlers': ['console'], 'level': 'DEBUG', 'propagate': False},
        'django.request': {'handlers': ['console'], 'level': 'DEBUG', 'propagate': False},
        'bx_django_utils': {'handlers': ['console'], 'level': 'DEBUG', 'propagate': False},
    },
}


if os.environ.get('RAISE_LOG_OUTPUT') in ('1', 'true'):
    # Raise an error on every uncaptured log message
    LOGGING['handlers']['raise_error'] = {
        'class': 'bx_py_utils.test_utils.log_utils.RaiseLogUsage',
    }
    for logger_cfg in LOGGING['loggers'].values():
        logger_cfg['handlers'] = ['raise_error']


# Playwright browser tests
# ----------------------------------------------------------------------------
# Avoid django.core.exceptions.SynchronousOnlyOperation. Playwright uses an event loop,
# even when using the sync API. Django only checks whether _any_ event loop is running,
# but not if _itself_ is running in an even loop.
# see https://github.com/microsoft/playwright-python/issues/439#issuecomment-763339612.
os.environ.setdefault('DJANGO_ALLOW_ASYNC_UNSAFE', '1')



# Generated by Django 3.1.4 on 2020-12-04 10:54

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ApproveTestModel',
            fields=[
                ('create_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.create_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.create_dt.verbose_name')),
                ('update_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.update_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.update_dt.verbose_name')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='BaseApproveModel.pk.help_text', primary_key=True, serialize=False, verbose_name='BaseApproveModel.pk.verbose_name')),
                ('is_draft', models.BooleanField(db_index=True, default=True, editable=False, help_text='BaseApproveModel.is_draft.help_text', verbose_name='BaseApproveModel.is_draft.verbose_name')),
                ('blocked', models.BooleanField(default=False, help_text='BaseApproveModel.blocked.help_text', verbose_name='BaseApproveModel.blocked.verbose_name')),
                ('ready_to_approve', models.BooleanField(blank=True, default=False, help_text='BaseApproveWorkflowModel.ready_to_approve.help_text', verbose_name='BaseApproveWorkflowModel.ready_to_approve.verbose_name')),
                ('name', models.CharField(help_text='This is always needed field', max_length=128, verbose_name='The Name')),
                ('title', models.CharField(blank=True, help_text='This is only needed to approve this entry', max_length=128, verbose_name='A Title')),
                ('approved', models.OneToOneField(editable=False, help_text='BaseApproveModel.approved.help_text', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='draft', to='approve_workflow_test_app.approvetestmodel', verbose_name='BaseApproveModel.approved.verbose_name')),
            ],
            options={
                'abstract': False,
                'default_permissions': ('add', 'change', 'delete', 'view', 'approve'),
            },
        ),
        migrations.CreateModel(
            name='RelatedApproveTestModel',
            fields=[
                ('create_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.create_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.create_dt.verbose_name')),
                ('update_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.update_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.update_dt.verbose_name')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='BaseApproveModel.pk.help_text', primary_key=True, serialize=False, verbose_name='BaseApproveModel.pk.verbose_name')),
                ('is_draft', models.BooleanField(db_index=True, default=True, editable=False, help_text='BaseApproveModel.is_draft.help_text', verbose_name='BaseApproveModel.is_draft.verbose_name')),
                ('blocked', models.BooleanField(default=False, help_text='BaseApproveModel.blocked.help_text', verbose_name='BaseApproveModel.blocked.verbose_name')),
                ('relation_name', models.CharField(help_text='This is always needed field', max_length=128, verbose_name='The Name')),
                ('relation_title', models.CharField(blank=True, help_text='This is only needed to approve this entry', max_length=128, null=True, verbose_name='A Title')),
                ('approved', models.OneToOneField(editable=False, help_text='BaseApproveModel.approved.help_text', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='draft', to='approve_workflow_test_app.relatedapprovetestmodel', verbose_name='BaseApproveModel.approved.verbose_name')),
                ('main_entry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='relations', to='approve_workflow_test_app.approvetestmodel')),
            ],
            options={
                'abstract': False,
                'default_permissions': ('add', 'change', 'delete', 'view', 'approve'),
            },
        ),
    ]



"""
    Helper to publish this Project to PyPi
"""

from pathlib import Path

from manageprojects.utilities.publish import publish_package
from manageprojects.utilities.subprocess_utils import verbose_check_call

import bx_django_utils


PACKAGE_ROOT = Path(bx_django_utils.__file__).parent.parent


def publish():
    """
    Publish to PyPi
    Call this via:
        $ poetry run publish
    """
    verbose_check_call('make', 'test')  # don't publish if tests fail
    verbose_check_call('make', 'fix-code-style')  # don't publish if code style wrong

    publish_package(
        module=bx_django_utils,
        package_path=PACKAGE_ROOT,
    )



import logging
from collections.abc import Iterable
from typing import Optional

from django.core.cache import cache

from bx_django_utils.feature_flags.exceptions import NotUniqueFlag
from bx_django_utils.feature_flags.models import FeatureFlagModel
from bx_django_utils.feature_flags.state import State
from bx_django_utils.feature_flags.utils import validate_cache_key
from bx_django_utils.models.manipulate import create_or_update2


logger = logging.getLogger(__name__)


class FeatureFlag:
    """
    A feature flag that persistent the state into django cache/database.
    """

    registry = {}

    def __init__(
        self,
        *,
        cache_key: str,
        human_name: str,
        initial_enabled: bool,
        description: Optional[str] = None,
        cache_key_prefix: str = 'feature-flags',
    ):
        self.human_name = human_name
        self.description = description

        validate_cache_key(cache_key)
        validate_cache_key(cache_key_prefix)
        self.cache_key = f'{cache_key_prefix}-{cache_key}'
        validate_cache_key(self.cache_key)  # Double check ;)

        if self.cache_key in self.registry:
            raise NotUniqueFlag(f'Cache key {self.cache_key !r} already registered!')

        self.registry[self.cache_key] = self

        if initial_enabled:
            self.initial_state = State.ENABLED
        else:
            self.initial_state = State.DISABLED

    def enable(self) -> bool:
        return self.set_state(new_state=State.ENABLED)

    def disable(self) -> bool:
        return self.set_state(new_state=State.DISABLED)

    def set_state(self, new_state: State) -> bool:
        """store state to database + cache and return it"""
        assert isinstance(
            new_state, State
        ), f'Given {new_state!r} (type: {type(new_state).__name__}) is not a State object!'

        state = self._add2cache(state_value=new_state.value)
        create_or_update2(
            ModelClass=FeatureFlagModel,
            lookup={'cache_key': self.cache_key},
            state=new_state,
        )
        return state

    def _add2cache(self, state_value: int) -> bool:
        """store state to cache and return it"""
        assert state_value in (0, 1), f'Unknown cache value: {state_value!r}'
        cache.set(self.cache_key, state_value, timeout=None)  # set forever
        return bool(state_value)

    @property
    def is_enabled(self) -> bool:
        try:
            raw_value = cache.get(self.cache_key)
        except Exception as err:
            # Will only happen, if cache backend is down
            logger.exception('Get cache key %r failed: %s', self.cache_key, err)
            raw_value = None  # Use the initial state

        if raw_value is None:
            # Not set, yet or cache was cleared -> Try to get the state from database
            instance = (
                FeatureFlagModel.objects.filter(cache_key=self.cache_key).values('state').first()
            )
            if instance:
                # We found the flag in database -> store state in the cache
                # and return the current state
                state_value = instance['state']
                return self._add2cache(state_value=state_value)
            else:
                # It's not in the cache and database -> store initial state in cache and database
                # and return the current state
                return self.set_state(new_state=self.initial_state)
        else:
            assert raw_value in (0, 1), f'Unknown cache value: {raw_value!r}'
            return bool(raw_value)

    @property
    def state(self):
        if self.is_enabled:
            return State.ENABLED
        else:
            return State.DISABLED

    @property
    def opposite_state(self):
        """
        returns the opposite of current "state".
        Usefull to toggle the state via set_state()
        """
        if self.is_enabled:
            return State.DISABLED
        else:
            return State.ENABLED

    @classmethod
    def values(cls) -> Iterable["FeatureFlag"]:
        yield from cls.registry.values()

    @classmethod
    def get_by_cache_key(cls, cache_key) -> "FeatureFlag":
        return cls.registry[cache_key]



from bx_py_utils.test_utils.unittest_utils import BaseDocTests

import bx_django_utils
import bx_django_utils_tests


class DocTests(BaseDocTests):
    def test_doctests(self):
        self.run_doctests(
            modules=(bx_django_utils, bx_django_utils_tests),
        )



import logging
from unittest.mock import patch

from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.test import TestCase

from bx_django_utils.feature_flags import data_classes
from bx_django_utils.feature_flags.data_classes import FeatureFlag
from bx_django_utils.feature_flags.exceptions import NotUniqueFlag
from bx_django_utils.feature_flags.state import State
from bx_django_utils.feature_flags.test_utils import FeatureFlagTestCaseMixin, get_feature_flag_db_info
from bx_django_utils.feature_flags.utils import validate_cache_key
from bx_django_utils.test_utils.assert_queries import AssertQueries


class FeatureFlagsTestCase(FeatureFlagTestCaseMixin, TestCase):
    """"""  # noqa - Don't add to README

    def setUp(self):
        super().setUp()

        FeatureFlag.registry.clear()
        self.initial_enabled_test_flag = FeatureFlag(
            cache_key='test-initial_enabled',
            human_name='A test flag that is initial "enabled"',
            initial_enabled=True,
        )
        self.initial_disabled_test_flag = FeatureFlag(
            cache_key='test-initial_disabled',
            human_name='A test flag that is initial "disabled"',
            initial_enabled=False,
        )

    def test_cache_key_validation(self):
        with self.assertRaisesMessage(ValidationError, 'cause errors if used with memcached'):
            validate_cache_key('Bam bam!')

        with self.assertRaisesMessage(ValidationError, 'cause errors if used with memcached'):
            FeatureFlag(
                cache_key='Bam bam!',
                human_name='Foo Bar',
                initial_enabled=True,
            )

    def test_uniqueness(self):
        FeatureFlag(
            cache_key='test',
            human_name='Test Uniqueness 1',
            initial_enabled=True,
        )
        with self.assertRaisesMessage(
            NotUniqueFlag, "Cache key 'feature-flags-test' already registered!"
        ):
            FeatureFlag(
                cache_key='test',
                human_name='Test Uniqueness 2',
                initial_enabled=True,
            )

    def test_basic(self):
        test_keys = {'feature-flags-test-initial_enabled', 'feature-flags-test-initial_disabled'}
        self.assertEqual(FeatureFlag.registry.keys(), test_keys)

        seen_keys = set()
        for instance in FeatureFlag.values():
            self.assertIsInstance(instance, FeatureFlag)
            seen_keys.add(instance.cache_key)
        self.assertEqual(seen_keys, test_keys)

        # Inistal state is enabled?
        self.assertEqual(self.initial_enabled_test_flag.state, State.ENABLED)
        self.assertEqual(self.initial_enabled_test_flag.opposite_state, State.DISABLED)
        self.assertTrue(self.initial_enabled_test_flag.is_enabled)
        # Disable
        self.initial_enabled_test_flag.set_state(State.DISABLED)
        # is disabled?
        self.assertFalse(self.initial_enabled_test_flag.is_enabled)
        self.assertEqual(self.initial_enabled_test_flag.state, State.DISABLED)
        self.assertEqual(self.initial_enabled_test_flag.opposite_state, State.ENABLED)
        # enable via set_state:
        self.initial_enabled_test_flag.set_state(State.ENABLED)
        # is enabled?
        self.assertEqual(self.initial_enabled_test_flag.state, State.ENABLED)
        self.assertEqual(self.initial_enabled_test_flag.opposite_state, State.DISABLED)
        self.assertTrue(self.initial_enabled_test_flag.is_enabled)

        # Just check the other initial state:
        self.assertEqual(self.initial_disabled_test_flag.state, State.DISABLED)
        self.assertEqual(self.initial_disabled_test_flag.opposite_state, State.ENABLED)
        self.assertFalse(self.initial_disabled_test_flag.is_enabled)

    def test_initial_state(self):
        self.assertEqual(self.initial_enabled_test_flag.state, State.ENABLED)
        self.assertTrue(self.initial_enabled_test_flag.is_enabled)

        self.assertEqual(self.initial_disabled_test_flag.state, State.DISABLED)
        self.assertFalse(self.initial_disabled_test_flag.is_enabled)

    def test_set_state(self):
        self.assertEqual(self.initial_enabled_test_flag.state, State.ENABLED)
        self.initial_enabled_test_flag.set_state(State.DISABLED)
        self.assertEqual(self.initial_enabled_test_flag.state, State.DISABLED)
        self.initial_enabled_test_flag.set_state(State.ENABLED)
        self.assertEqual(self.initial_enabled_test_flag.state, State.ENABLED)

    def test_enable_disable(self):
        self.assertEqual(self.initial_enabled_test_flag.state, State.ENABLED)
        self.initial_enabled_test_flag.disable()
        self.assertEqual(self.initial_enabled_test_flag.state, State.DISABLED)
        self.initial_enabled_test_flag.enable()
        self.assertEqual(self.initial_enabled_test_flag.state, State.ENABLED)

    def test_redis_down(self):
        # Don't crash if "cache.get()" raise an error, e.g.: If redis is down
        self.assertTrue(self.initial_enabled_test_flag.is_enabled)
        with patch.object(
            data_classes.cache, 'get', side_effect=Exception('Cache down')
        ), self.assertLogs('bx_django_utils', level=logging.ERROR) as logs:
            self.assertTrue(self.initial_enabled_test_flag.is_enabled)

        output = '\n'.join(logs.output)
        self.assertIn(
            'Get cache key \'feature-flags-test-initial_enabled\' failed: Cache down', output
        )

    def test_cache_usage(self):
        # Test if cache works, in current test setup:
        self.assertIsNone(cache.get('foobar'))
        cache.set('foobar', 1)
        self.assertEqual(cache.get('foobar'), 1)

        # Test if FeatureFlag used the cache:
        CACHE_KEY = 'feature-flags-foo'

        # Create the flag:

        flag = FeatureFlag(cache_key='foo', human_name='foo', initial_enabled=False)
        self.assertEqual(flag.cache_key, CACHE_KEY)

        # Initializing will *not* store anything:
        self.assertIsNone(cache.get(CACHE_KEY))
        self.assertEqual(get_feature_flag_db_info(), {})

        # Fetch stage -> Store in cache + database:
        with AssertQueries() as queries:
            self.assertFalse(flag.is_enabled)  # Initial: disabled

        # First time init takes 3 queries:
        query_count = 1  # Lookup value
        query_count += 1  # create_or_update2() second lookup
        query_count += 1  # create_or_update2() insert
        queries.assert_queries(
            table_counts={'feature_flags_featureflagmodel': query_count},
            double_tables=False,
        )

        # Now it's in the cache:
        self.assertIs(cache.get(CACHE_KEY), 0)
        # ...and in the database:
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-foo': 0})

        # Check flag again should only hit the cache:
        with AssertQueries() as queries:
            self.assertFalse(flag.is_enabled)  # Still disabled?
        queries.assert_queries(table_counts={})  # No Queries made?

        # Change the flag should store the new stage in cache + database:
        with AssertQueries() as queries:
            flag.enable()
        query_count = 1  # create_or_update2() lookup
        query_count += 1  # create_or_update2() change values
        queries.assert_queries(
            table_counts={'feature_flags_featureflagmodel': query_count},
            double_tables=False,
        )
        # Changed in the cache:
        self.assertIs(cache.get(CACHE_KEY), 1)
        # ...and in the database:
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-foo': 1})

        # Query will only use the cache:
        with AssertQueries() as queries:
            self.assertTrue(flag.is_enabled)  # Now it's enabled
        queries.assert_queries(table_counts={})  # No Queries made?



import traceback

from django.db.models.signals import post_init


class CleanMock:
    """
    Track if full_clean() was called.
    """

    def __init__(self, instance):
        self.instance = instance  # the model instance that should be tracked

        # Replace the full_clean() function to track the call of it:
        self.origin_clean = instance.full_clean
        instance.full_clean = self.full_clean

        self.clean_called = False  # Was full_clean() called?

        # Helpful to find the code part that missed the full_clean() call:
        self.stack = traceback.extract_stack(f=None, limit=None)

    def full_clean(self, *args, **kwargs):
        self.clean_called = True
        return self.origin_clean(*args, **kwargs)

    def model_pkg_name(self):
        opts = self.instance._meta
        return f'{opts.app_label}.{opts.object_name}'

    def __repr__(self):
        return f'<CleanMock {self.model_pkg_name()}>'


class AssertModelCleanCalled:
    """
    Context manager for assert that full_clean() was called for every model instance.
    """

    def __init__(self):
        self.clean_mocks = []
        self.called_cleans = []
        self.missing_cleans = []

    def post_init_handler(self, signal, sender, instance, **kwargs):
        self.clean_mocks.append(CleanMock(instance))

    def __enter__(self):
        post_init.connect(self.post_init_handler)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        post_init.disconnect(self.post_init_handler)
        for clean_mock in self.clean_mocks:
            if clean_mock.clean_called:
                self.called_cleans.append(clean_mock)
            else:
                self.missing_cleans.append(clean_mock)

    def assert_no_missing_cleans(self):
        """
        Assert that there are no missing full_clean() calls.
        Give a very verbose error message with a stack trace
        so it should be easy to find the related code part.
        """
        if not self.missing_cleans:
            # There are no missing full_clean() calls -> everything is okay
            return

        msg = [
            f'There are {len(self.missing_cleans)} missing clean calls:'
        ]
        seen_stacks = []
        for missing_clean in self.missing_cleans:

            stack_info = ''.join(traceback.format_list(missing_clean.stack))
            if stack_info in seen_stacks:
                # Skip same stack traces
                continue
            seen_stacks.append(stack_info)

            msg.append(
                f'{missing_clean.model_pkg_name()}: {stack_info}'
            )

        raise AssertionError('\n'.join(msg))



import logging
import uuid

from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.utils.translation import gettext_lazy as _

from bx_django_utils.models.timetracking import TimetrackingBaseModel


logger = logging.getLogger(__name__)


class BaseApproveModel(TimetrackingBaseModel):
    """
    Base model class for approve models *and* this relation models.
    """
    id = models.UUIDField(
        # The internal primary key
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
        verbose_name=_('BaseApproveModel.pk.verbose_name'),
        help_text=_('BaseApproveModel.pk.help_text')
    )
    is_draft = models.BooleanField(
        # Marks this entry as "draft"
        # Only drafts can be "approve"
        default=True,
        editable=False,
        db_index=True,
        verbose_name=_('BaseApproveModel.is_draft.verbose_name'),
        help_text=_('BaseApproveModel.is_draft.help_text')
    )
    approved = models.OneToOneField(
        # Relation to the "approved" instance
        # Set by self.approve() to the draft instance
        'self',
        on_delete=models.SET_NULL,
        related_name='draft',
        null=True,
        editable=False,
        verbose_name=_('BaseApproveModel.approved.verbose_name'),
        help_text=_('BaseApproveModel.approved.help_text')
    )
    blocked = models.BooleanField(
        default=False,
        verbose_name=_('BaseApproveModel.blocked.verbose_name'),
        help_text=_('BaseApproveModel.blocked.help_text')
    )

    # Fields are required for *approve* instances (should be used in forms):
    REQUIRED_FIELDS_PUBLIC = ()

    def is_approved(self):
        if not self.is_draft:
            # We are the "approved" instance
            return True
        return self.approved_id is not None
    is_approved.short_description = _('Approved')
    is_approved.boolean = True

    def get_missing_field_info(self, ignore_fields=None, extra_fields=None):
        needed_fields = set(self.REQUIRED_FIELDS_PUBLIC)

        if extra_fields:
            needed_fields |= set(extra_fields)

        if ignore_fields:
            needed_fields -= set(ignore_fields)

        missing = []
        for field_name in sorted(needed_fields):
            field = self._meta.get_field(field_name)

            if field.remote_field:
                attname = field_name
                verbose_name = field.name
            else:
                attname = field.get_attname()
                verbose_name = field.verbose_name

            value = getattr(self, attname)
            if field.remote_field and not isinstance(field.remote_field, models.ManyToOneRel):
                # Exists a many2many
                value = value.exists()

            if not value:
                missing.append(
                    (attname, verbose_name, value)
                )
        return missing

    def full_clean(self, **kwargs):
        super().full_clean(**kwargs)
        if not self.is_draft:
            missing = self.get_missing_field_info()
            errors = {}
            for attname, verbose_name, value in missing:
                errors[attname] = ValidationError(
                    _('This field cannot be blank.'),
                    code='blank'
                )
            if errors:
                raise ValidationError(errors)

    def _copy_normal_fields(self, approve_instance):
        """
        Just assign all normal field values from draft to approve version.
        """
        raise NotImplementedError

    def _copy_relations(self, approve_instance):
        """
        Assign relations from draft to approve version.
        """
        raise NotImplementedError

    def approve(self, **values):
        """
        1. validate that draft in complete filled
        2. Create/update a approve instance with all data from the draft
        """
        logger.debug('Approve %s', self)

        if not self.is_draft:
            raise ValidationError('Only drafts can be approve!')

        if self.blocked:
            raise ValidationError('Blocked entries can not approved!')

        # Check validation before creating the approve version
        self.full_clean()

        with transaction.atomic():
            if self.approved is None:
                approve_instance = self.__class__(
                    is_draft=False,
                )
            else:
                approve_instance = self.approved

            assert approve_instance.is_draft is False

            for key, value in values.items():
                setattr(approve_instance, key, value)

            # "normal" field values can be assigned before a new approve instance
            # has been saved and has a primary key
            self._copy_normal_fields(approve_instance)

            # Transfer date times:
            approve_instance.create_dt = self.create_dt
            approve_instance.update_dt = self.update_dt

            # Save the instance to set a primary key
            # PK is needed to assign relations
            approve_instance.save(
                # Maybe relations are needed -> skip validation
                # validation will be done later, see below
                call_clean=False,
                update_dt=False  # Keep date time from draft
            )

            # Assign all relation data
            self._copy_relations(approve_instance)

            # If fields are missing or validation failed
            # -> raise and error and abort the transaction
            approve_instance.full_clean()

            if self.approved_id is not approve_instance.pk:
                # Save approve instance to draft version
                self.approved_id = approve_instance.pk
                self.save(
                    update_fields=('approved_id',),
                    update_dt=False  # Keep date time from draft
                )

        logger.info('%s was approved to %s', self, approve_instance)
        return approve_instance

    def save(self, call_clean=True, **kwargs):
        if call_clean:
            # The approve version should be validated
            self.full_clean()

        return super().save(**kwargs)

    def __repr__(self):
        if self.is_draft:
            info = 'draft'
        else:
            info = 'approved'
        return f'<{self.__class__.__name__} pk:{self.pk} ({info})>'

    def __str__(self):
        return self.__repr__()

    class Meta:
        default_permissions = ('add', 'change', 'delete', 'view', 'approve')  # Add 'approve'
        abstract = True


class BaseApproveWorkflowModel(BaseApproveModel):
    """
    Base model for approve workflow models.
    Don't used this for approve relation models!
    """
    ready_to_approve = models.BooleanField(
        # Marks when it's ready to approve
        default=False,
        null=False,
        blank=True,
        verbose_name=_('BaseApproveWorkflowModel.ready_to_approve.verbose_name'),
        help_text=_('BaseApproveWorkflowModel.ready_to_approve.help_text')
    )

    def approve(self, **values):
        if not self.is_draft:
            raise ValidationError('Only drafts can be approve!')

        if not self.ready_to_approve:
            logger.warning('Not ready to approve: pk:%r model:%s', self.pk, type(self).__name__)

        with transaction.atomic():
            # Remove the "ready" flag for next edit:
            self.ready_to_approve = False
            self.save(
                update_fields=('ready_to_approve',),
                update_dt=False  # Keep date time from draft
            )

            approve_instance = super().approve(**values)

            # self._copy_normal_fields() should not copy the "ready" flag:
            assert not approve_instance.ready_to_approve

        return approve_instance

    def __repr__(self):
        if self.is_draft:
            if self.ready_to_approve:
                info = 'draft, ready to approve'
            else:
                info = 'draft, not ready'
        else:
            info = 'approved'
        return f'<{self.__class__.__name__} pk:{self.pk} ({info})>'

    def __str__(self):
        return self.__repr__()

    class Meta(BaseApproveModel.Meta):
        abstract = True



# Generated by Django 4.1.5 on 2023-02-07 15:29

from django.db import migrations, models

import bx_django_utils.translation


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0002_storesavemodel"),
    ]

    operations = [
        migrations.CreateModel(
            name="TranslatedModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "translated",
                    bx_django_utils.translation.TranslationField(["de-de", "en-us", "es"]),
                ),
                (
                    "translated_multiline",
                    bx_django_utils.translation.TranslationField(["de-de", "en-us", "es"], blank=True),
                ),
            ],
        ),
    ]



import debug_toolbar
from django.contrib import admin
from django.urls import include, path

from bx_django_utils.admin_extra_views.registry import extra_view_registry
from bx_django_utils.admin_extra_views.views import Redirect2AdminExtraView
from bx_django_utils_tests.test_app.admin_views import DemoView3


urlpatterns = [
    path('admin/', include(extra_view_registry.get_urls())),
    path('old_demo_3/', Redirect2AdminExtraView.as_view(admin_view=DemoView3)),
    path('admin/', admin.site.urls),
    path('', include('bx_django_utils_tests.test_app.urls')),
    path('__debug__/', include(debug_toolbar.urls)),
]



from unittest import mock

from bx_py_utils.test_utils.datetime import parse_dt
from django.core.exceptions import ValidationError
from django.core.validators import validate_slug
from django.db.utils import IntegrityError
from django.test import TestCase
from django.utils import timezone

from bx_django_utils.models.manipulate import (
    STORE_BEHAVIOR_IGNORE,
    STORE_BEHAVIOR_SET_IF_EMPTY,
    STORE_BEHAVIOR_SKIP_EMPTY,
    CreateOrUpdateResult,
    FieldUpdate,
    InvalidStoreBehavior,
    create,
    create_or_update,
    create_or_update2,
)
from bx_django_utils.test_utils.datetime import MockDatetimeGenerator
from bx_django_utils.test_utils.model_clean_assert import AssertModelCleanCalled
from bx_django_utils_tests.test_app.models import CreateOrUpdateTestModel, StoreSaveModel, TimetrackingTestModel


class ModelManipulateTestCase(TestCase):
    def test_deprecated_create_or_update(self):
        with self.assertWarns(DeprecationWarning):
            instance, created, updated_fields = create_or_update(
                ModelClass=CreateOrUpdateTestModel, name='foo', slug='bar'
            )
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.name == 'foo'
            assert created is True
            assert updated_fields is None  # None and not []

    @mock.patch.object(timezone, 'now', MockDatetimeGenerator())
    def test_create_or_update2(self):

        # create a new entry:

        with AssertModelCleanCalled() as cm:
            result = create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup={'id': 1},
                name='First entry',
                slug='first'
            )
            self.assertIsInstance(result, CreateOrUpdateResult)
            instance = result.instance
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.id == 1
            assert instance.name == 'First entry'
            assert instance.slug == 'first'
            assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
            assert instance.update_dt == parse_dt('2001-01-01T00:00:00+0000')
            assert result.created is True
            assert result.ignored_fields == []
            assert result.not_overwritten_fields == []
            # Not set if new instance created:
            self.assertEqual(result.updated_fields, [])
            self.assertEqual(result.update_info, [])
        cm.assert_no_missing_cleans()

        # Change only 'slug'

        with AssertModelCleanCalled() as cm:
            result = create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup={'id': 1},
                name='First entry',
                slug='change-value'
            )
            instance = result.instance
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.id == 1
            assert instance.name == 'First entry'
            assert instance.slug == 'change-value'
            assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')  # not changed!
            assert instance.update_dt == parse_dt('2002-01-01T00:00:00+0000')
            assert result.created is False
            assert result.ignored_fields == []
            assert result.not_overwritten_fields == []
            self.assertEqual(result.updated_fields, ['slug'])
            self.assertEqual(
                result.update_info, [FieldUpdate(field_name='slug', old_value='first', new_value='change-value')]
            )
        cm.assert_no_missing_cleans()

        # Change 'name' and 'slug':

        with AssertModelCleanCalled() as cm:
            result = create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup={'id': 1},
                name='New name !',
                slug='new-slug'
            )
            instance = result.instance
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.id == 1
            assert instance.name == 'New name !'
            assert instance.slug == 'new-slug'
            assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')  # not changed!
            assert instance.update_dt == parse_dt('2003-01-01T00:00:00+0000')
            assert result.created is False
            assert result.ignored_fields == []
            assert result.not_overwritten_fields == []
            self.assertEqual(result.updated_fields, ['name', 'slug'])
            self.assertEqual(
                result.update_info,
                [
                    FieldUpdate(field_name='name', old_value='First entry', new_value='New name !'),
                    FieldUpdate(field_name='slug', old_value='change-value', new_value='new-slug'),
                ],
            )
        cm.assert_no_missing_cleans()

        # Nothing changed:

        result = create_or_update2(
            ModelClass=CreateOrUpdateTestModel,
            lookup={'id': 1},
            name='New name !',
            slug='new-slug'
        )
        instance = result.instance
        assert isinstance(instance, CreateOrUpdateTestModel)
        assert instance.id == 1
        assert instance.name == 'New name !'
        assert instance.slug == 'new-slug'
        assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert instance.update_dt == parse_dt('2003-01-01T00:00:00+0000')  # not changed!
        assert result.created is False
        assert result.ignored_fields == []
        assert result.not_overwritten_fields == []
        self.assertEqual(result.updated_fields, [])
        self.assertEqual(result.update_info, [])

    def test_non_valid(self):
        msg = str(validate_slug.message)
        with self.assertRaisesMessage(ValidationError, msg):
            create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup={'id': 1},
                name='foo',
                slug='this is no Slug !'
            )

        # Update existing entry with non-valid values should also not work:

        CreateOrUpdateTestModel(id=1, name='foo', slug='bar')
        with self.assertRaisesMessage(ValidationError, msg):
            create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup={'id': 1},
                name='foo',
                slug='this is no Slug !'
            )

    def test_disable_full_clean(self):
        # Create a new entry without "full_clean()" call:
        with AssertModelCleanCalled() as cm:
            result = create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup={'id': 1},
                call_full_clean=False,
                slug='This is not a valid slug!'
            )
            instance = result.instance
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.id == 1
            assert instance.slug == 'This is not a valid slug!'
            assert result.created is True
            assert result.updated_fields == []
        assert cm.called_cleans == []
        assert len(cm.missing_cleans) == 1

        # Change existing without "full_clean()" call:
        with AssertModelCleanCalled() as cm:
            result = create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup={'id': 1},
                call_full_clean=False,
                slug='Also no valid slug!'
            )
            instance = result.instance
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.id == 1
            assert instance.slug == 'Also no valid slug!'
            assert result.created is False
            assert result.updated_fields == ['slug']
        assert cm.called_cleans == []
        assert len(cm.missing_cleans) == 1

    @mock.patch.object(timezone, 'now', MockDatetimeGenerator())
    def test_create_or_update_without_lookup(self):
        # create a new entry:

        with AssertModelCleanCalled() as cm:
            result = create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                lookup=None,
                name='First entry',
                slug='first'
            )
            instance = result.instance
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.pk is not None
            assert instance.name == 'First entry'
            assert instance.slug == 'first'
            assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
            assert instance.update_dt == parse_dt('2001-01-01T00:00:00+0000')
            assert result.created is True
            assert result.updated_fields == []
        cm.assert_no_missing_cleans()

    @mock.patch.object(timezone, 'now', MockDatetimeGenerator())
    def test_create(self):
        # create a new entry:

        with AssertModelCleanCalled() as cm:
            instance = create(
                ModelClass=CreateOrUpdateTestModel,
                name='First entry',
                slug='first'
            )
            assert isinstance(instance, CreateOrUpdateTestModel)
            assert instance.pk is not None
            assert instance.name == 'First entry'
            assert instance.slug == 'first'
            assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
            assert instance.update_dt == parse_dt('2001-01-01T00:00:00+0000')
        cm.assert_no_missing_cleans()

        # Cannot create an already existing model
        with self.assertRaises(IntegrityError):
            create(
                ModelClass=CreateOrUpdateTestModel,
                id=instance.id,
                name='second create',
                slug='second'
            )

    @mock.patch.object(timezone, 'now', MockDatetimeGenerator())
    def test_store_behavior(self):
        test_relation1 = TimetrackingTestModel(
            create_dt=parse_dt('2002-02-02T00:00:00+0000'),
            update_dt=parse_dt('2003-03-03T00:00:00+0000')
        )
        test_relation1.save(update_dt=False)

        test_relation2 = TimetrackingTestModel(
            create_dt=parse_dt('2004-04-04T00:00:00+0000'),
            update_dt=parse_dt('2005-05-05T00:00:00+0000')
        )
        test_relation2.save(update_dt=False)

        # Create object and respect "store_behavior"

        result = create_or_update2(
            ModelClass=CreateOrUpdateTestModel,
            lookup=None,  # force create object!
            store_behavior={
                # 'name' is missing here -> normal behavior: overwrite existing values
                'slug': STORE_BEHAVIOR_SET_IF_EMPTY,
                'many2one_rel': STORE_BEHAVIOR_SET_IF_EMPTY,
                'blank_field': STORE_BEHAVIOR_IGNORE,
                'null_field': STORE_BEHAVIOR_IGNORE,
            },
            name='name1',
            slug='slug1',
            many2one_rel=test_relation1,
            blank_field='ignored',
            null_field='ignored',
        )
        assert result.created is True
        assert result.updated_fields == []  # Object created!
        assert sorted(result.ignored_fields) == ['blank_field', 'null_field']
        assert result.not_overwritten_fields == []
        assert result.skip_empty_values == []
        instance = result.instance
        assert instance.name == 'name1'
        assert instance.slug == 'slug1'
        assert instance.many2one_rel.create_dt == parse_dt('2002-02-02T00:00:00+0000')
        assert instance.many2one_rel.update_dt == parse_dt('2003-03-03T00:00:00+0000')
        assert instance.blank_field == ''
        assert instance.null_field is None
        assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert instance.update_dt == parse_dt('2001-01-01T00:00:00+0000')

        # Update existing instance

        result = create_or_update2(
            ModelClass=CreateOrUpdateTestModel,
            lookup={'pk': instance.pk},
            store_behavior={
                # 'name' is missing here -> normal behavior: overwrite existing values
                'slug': STORE_BEHAVIOR_SET_IF_EMPTY,
                'many2one_rel': STORE_BEHAVIOR_SKIP_EMPTY,  # given relation is not empty
                'blank_field': STORE_BEHAVIOR_SET_IF_EMPTY,
                'null_field': STORE_BEHAVIOR_SET_IF_EMPTY,
            },
            name='name2',
            slug='not-overwritten',
            many2one_rel=test_relation2,
            blank_field='set blank field 1',
            null_field='set null field 1',
        )
        instance = result.instance
        assert result.created is False
        assert instance.name == 'name2'
        assert instance.slug == 'slug1'
        assert instance.many2one_rel.create_dt == parse_dt('2004-04-04T00:00:00+0000')  # updated
        assert instance.many2one_rel.update_dt == parse_dt('2005-05-05T00:00:00+0000')  # updated
        assert instance.blank_field == 'set blank field 1'
        assert instance.null_field == 'set null field 1'
        assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert instance.update_dt == parse_dt('2002-01-01T00:00:00+0000')
        assert sorted(result.updated_fields) == [
            'blank_field', 'many2one_rel', 'name', 'null_field'
        ]
        assert result.ignored_fields == []
        assert result.not_overwritten_fields == ['slug']
        assert result.skip_empty_values == []

        # Skip empty values

        result = create_or_update2(
            ModelClass=CreateOrUpdateTestModel,
            lookup={'pk': instance.pk},
            store_behavior={
                'slug': STORE_BEHAVIOR_IGNORE,
                'many2one_rel': STORE_BEHAVIOR_SKIP_EMPTY,
                'blank_field': STORE_BEHAVIOR_SKIP_EMPTY,
                'null_field': STORE_BEHAVIOR_SKIP_EMPTY,
            },
            name='name3',
            slug='will-be-ignored',
            many2one_rel=None,
            blank_field='',  # a empty value
            null_field=None,  # a empty value
        )
        instance = result.instance
        assert result.created is False
        assert instance.name == 'name3'  # new name
        assert instance.slug == 'slug1'  # unchanged
        assert instance.many2one_rel.create_dt == parse_dt('2004-04-04T00:00:00+0000')  # unchanged
        assert instance.many2one_rel.update_dt == parse_dt('2005-05-05T00:00:00+0000')  # unchanged
        assert instance.blank_field == 'set blank field 1'  # unchanged
        assert instance.null_field == 'set null field 1'  # unchanged
        assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert instance.update_dt == parse_dt('2003-01-01T00:00:00+0000')
        assert result.updated_fields == ['name']
        assert result.ignored_fields == ['slug']
        assert result.not_overwritten_fields == []
        assert sorted(result.skip_empty_values) == [
            'blank_field', 'many2one_rel', 'null_field'
        ]

        # Store empty values

        result = create_or_update2(
            ModelClass=CreateOrUpdateTestModel,
            lookup={'pk': instance.pk},
            store_behavior={
                'name': STORE_BEHAVIOR_IGNORE,
                'slug': STORE_BEHAVIOR_IGNORE,
            },
            name='Not Overwritten !',
            # "slug" missing here, but can be set in "store_behavior"
            many2one_rel=None,  # can be set to "empty"
            blank_field='',  # can be set to "empty"
            null_field=None,  # can be set to "empty"
        )
        instance = result.instance
        assert result.created is False
        assert instance.name == 'name3'  # unchanged
        assert instance.slug == 'slug1'  # unchanged
        assert instance.many2one_rel is None
        assert instance.blank_field == ''
        assert instance.null_field is None
        assert instance.create_dt == parse_dt('2001-01-01T00:00:00+0000')
        assert instance.update_dt == parse_dt('2004-01-01T00:00:00+0000')
        assert sorted(result.updated_fields) == ['blank_field', 'many2one_rel', 'null_field']
        assert result.ignored_fields == ['name']
        assert result.not_overwritten_fields == []

        # We accept only existing field names in store_behavior:

        err_msg = (
            "store_behavior field name 'wrong' is not one of:"
            " ['blank_field', 'create_dt', 'id', 'many2one_rel',"
            " 'name', 'null_field', 'slug', 'update_dt']"
        )
        with self.assertRaisesMessage(InvalidStoreBehavior, err_msg):
            create_or_update2(
                ModelClass=CreateOrUpdateTestModel,
                store_behavior={
                    'name': STORE_BEHAVIOR_IGNORE,
                    'slug': STORE_BEHAVIOR_SET_IF_EMPTY,
                    # We check the field names:
                    'wrong': STORE_BEHAVIOR_IGNORE,
                },
            )

        assert CreateOrUpdateTestModel.objects.count() == 1

    @mock.patch.object(timezone, 'now', MockDatetimeGenerator())
    def test_save_kwargs(self):
        obj = create_or_update2(
            ModelClass=StoreSaveModel,
            name='foobar',
            save_kwargs={'arg': 'original'},
        ).instance
        assert obj.name == 'foobar'

        create_or_update2(
            ModelClass=StoreSaveModel,
            lookup={'pk': obj.pk},
            name='bazqux',
            save_kwargs={'other_arg': 'changed'},
        )
        obj.refresh_from_db()
        assert obj.name == 'bazqux'

        create_or_update2(
            ModelClass=StoreSaveModel,
            lookup={'pk': obj.pk},
            name='final',
            save_kwargs={},
        )
        obj.refresh_from_db()
        assert obj.name == 'final'

        assert obj._save_calls.saves == [
            {'arg': 'original'},
            {'other_arg': 'changed'},
            {},
        ]



from django.apps import AppConfig


class UserTimezoneAppConfig(AppConfig):
    """
    Django app to set the user local time zone.
    """

    name = 'bx_django_utils.user_timezone'
    verbose_name = "User Timezone"

    def ready(self):
        # Register system checks:
        import bx_django_utils.user_timezone.checks  # noqa



from django.db import connections
from django.db.backends.utils import CursorWrapper
from django.test import TestCase

from bx_django_utils.dbperf.cursor import RecordingCursorWrapper
from bx_django_utils.dbperf.query_recorder import SQLQueryRecorder
from bx_django_utils_tests.test_app.models import CreateOrUpdateTestModel


class SQLQueryRecorderTestCase(TestCase):
    def test_cursor_wrapping(self):
        self.assertIsInstance(connections['default'].cursor(), CursorWrapper)
        with SQLQueryRecorder(databases=self.databases):
            self.assertIsInstance(connections['default'].cursor(), RecordingCursorWrapper)
        self.assertIsInstance(connections['default'].cursor(), CursorWrapper)

    def test_results_not_accessible_during_run(self):
        with SQLQueryRecorder(databases=self.databases) as rec:
            CreateOrUpdateTestModel.objects.count()
            with self.assertRaises(AssertionError):
                _ = rec.results()  # accessed before exiting context

    def test_results(self):
        with SQLQueryRecorder(databases=self.databases) as rec:
            CreateOrUpdateTestModel.objects.count()

        res = rec.results()
        self.assertCountEqual(res, (
            'queries',
            'databases',
            'sql_time',
            'num_queries',
            'queries_similar',
            'queries_duplicated',
            'num_queries_similar',
            'num_queries_duplicated',
        ))
        self.assertEqual(len(res['queries']), 1)
        self.assertGreater(res['sql_time'], 0)
        self.assertEqual(res['num_queries'], 1)
        self.assertEqual(len(res['queries_similar']), 1)
        self.assertEqual(len(res['queries_duplicated']), 1)

    def test_results_aggregation(self):
        with SQLQueryRecorder(databases=self.databases) as rec:
            # do some queries, explanation see below
            CreateOrUpdateTestModel.objects.count()
            CreateOrUpdateTestModel.objects.count()
            try:
                CreateOrUpdateTestModel.objects.only('id').get(id=99)
            except CreateOrUpdateTestModel.DoesNotExist:
                pass  # we just want the query to happen, not interested in the result
            len(CreateOrUpdateTestModel.objects.filter(name__iexact='foo').only('id'))
            len(CreateOrUpdateTestModel.objects.filter(name__iexact='bar').only('id'))
            len(CreateOrUpdateTestModel.objects.filter(name__iexact='baz').only('id'))
            len(CreateOrUpdateTestModel.objects.filter(name__iexact='foo').only('id'))

        res = rec.results()
        self.assertEqual(len(res['queries']), 7)
        self.assertGreater(res['sql_time'], 0)
        self.assertEqual(res['num_queries'], 7)

        self.assertEqual(len(res['queries_similar']['default']), 2)
        self.assertEqual(len(res['queries_duplicated']['default']), 2)
        self.assertEqual(res['num_queries_similar'], 2)
        self.assertEqual(res['num_queries_duplicated'], 2)

        #
        # CreateOrUpdateTestModel.objects.count()
        # we ran it twice, it should show as both similar and duplicate
        # because it doesn't have any params
        #
        query = 'SELECT COUNT(*) AS "__count" FROM "test_app_createorupdatetestmodel"'
        self.assertEqual(res['queries_similar']['default'][query], 2)
        self.assertEqual(res['queries_duplicated']['default'][(query, '()')], 2)

        #
        # CreateOrUpdateTestModel.objects.only('id').first()
        # we ran it once, so it shouldn't appear in either aggregation dict
        #
        query = 'SELECT "test_app_createorupdatetestmodel"."id" FROM "test_app_createorupdatetestmodel" WHERE "test_app_createorupdatetestmodel"."id" = %s'  # noqa: E501
        self.assertNotIn(query, res['queries_similar']['default'])
        self.assertNotIn(
            (query, "('00000000000000000000000000000063',)"),
            res['queries_duplicated']['default'],
        )

        #
        # TonieCloudUser.objects.filter(email__iexact=...).only('id')
        # we ran it quadruply, twice with the same params
        # it should show as both similar and duplicated
        #
        query = 'SELECT "test_app_createorupdatetestmodel"."id" FROM "test_app_createorupdatetestmodel" WHERE "test_app_createorupdatetestmodel"."name" LIKE %s ESCAPE \'\\\''  # noqa: E501
        self.assertEqual(res['queries_similar']['default'][query], 4)
        self.assertEqual(res['queries_duplicated']['default'][(query, "('foo',)")], 2)
        self.assertNotIn((query, "('bar',)"), res['queries_duplicated']['default'])
        self.assertNotIn((query, "('baz',)"), res['queries_duplicated']['default'])



from django.http import QueryDict


def build_url_parameters(safe_=None, **params):
    """
    Return an encoded string of all given parameters.

    >>> build_url_parameters(foo=1, bar='<code>')
    'bar=%3Ccode%3E&foo=1'

    List and tuples can be also used, e.g.:

    >>> build_url_parameters(foo=(1,2), bar=[3,4])
    'bar=3&bar=4&foo=1&foo=2'

    `safe_` specifies characters which don't require quoting, e.g.:

    >>> build_url_parameters(next='/foo&bar/')
    'next=%2Ffoo%26bar%2F'

    >>> build_url_parameters(next='/foo&bar/', safe_='/')
    'next=/foo%26bar/'
    """
    query = QueryDict(mutable=True)
    for key, value in sorted(params.items()):
        if isinstance(value, (tuple, list)):
            query.setlist(key, value)
        else:
            query[key] = value
    return query.urlencode(safe=safe_)



import random

from django.test import TestCase

from bx_django_utils.feature_flags.data_classes import FeatureFlag
from bx_django_utils.feature_flags.test_utils import (
    FeatureFlagTestCaseMixin,
    get_feature_flag_cache_info,
    get_feature_flag_db_info,
)


TEST_CACHE_KEY = 'feature-flags-test_atomic'


class FeatureFlagTestCaseMixinTestCase(FeatureFlagTestCaseMixin, TestCase):
    """"""  # noqa - Don't add to README

    def _test_atomic(self):
        # We always start fresh: Only with existing entries:
        self.assertEqual(FeatureFlag.registry.keys(), {'feature-flags-foo', 'feature-flags-bar'})

        # Cache and DB empty?
        self.assertEqual(get_feature_flag_cache_info(), {})
        self.assertEqual(get_feature_flag_db_info(), {})

        # Add a new "test_atomic" entry in FeatureFlag.registry:
        self.assertNotIn(TEST_CACHE_KEY, FeatureFlag.registry)
        bar_feature_flag = FeatureFlag(cache_key='test_atomic', human_name='bar', initial_enabled=True)
        self.assertEqual(bar_feature_flag.cache_key, TEST_CACHE_KEY)
        self.assertIn(TEST_CACHE_KEY, FeatureFlag.registry)
        self.assertTrue(bar_feature_flag.is_enabled)  # Nothing persistent?
        bar_feature_flag.disable()  # Add persistent state in cache + database
        self.assertFalse(bar_feature_flag.is_enabled)

        # Only our "test_atomic" stored?
        self.assertEqual(get_feature_flag_cache_info(), {'feature-flags-test_atomic': 0})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-test_atomic': 0})

        # Add garbage to registry -> FeatureFlagTestCaseMixin should reset it:
        FeatureFlag.registry[random.randrange(1000)] = random.randrange(1000)

    def test_atomic1(self):
        self._test_atomic()

    def test_atomic2(self):
        self._test_atomic()



import zoneinfo

from bx_py_utils.test_utils.datetime import parse_dt
from django.test import SimpleTestCase, override_settings
from django.utils import timezone, translation

from bx_django_utils.test_utils.html_assertion import assert_html_snapshot
from bx_django_utils.user_timezone.humanize import human_timezone_datetime


@override_settings(VISIBLE_TIMEZONES=['Europe/Berlin', 'America/Los_Angeles'])
class HumanizeTestCase(SimpleTestCase):
    def test_human_timezone_datetime(self):
        with timezone.override(zoneinfo.ZoneInfo('UTC')), translation.override('en'):
            html = human_timezone_datetime(dt=parse_dt('2000-01-01T00:00:00+0000'))
        self.assertIn('Europe/Berlin: Jan. 1, 2000, 1 a.m.', html)
        self.assertIn('America/Los_Angeles: Dec. 31, 1999, 4 p.m.', html)
        self.assertIn('01/01/2000 midnight', html)
        self.assertInHTML('<small>(UTC)</small>', html)
        assert_html_snapshot(got=html, validate=True)

        with timezone.override(zoneinfo.ZoneInfo('Europe/Berlin')), translation.override('de-de'):
            html = human_timezone_datetime(dt=parse_dt('2000-01-01T00:00:00+0000'))
        self.assertIn('Europe/Berlin: 1. Januar 2000 01:00', html)
        self.assertIn('America/Los_Angeles: 31. Dezember 1999 16:00', html)
        self.assertIn('01.01.2000 01:00', html)
        self.assertInHTML('<small>(Europe/Berlin)</small>', html)
        assert_html_snapshot(got=html, validate=True)

        # No <span title=""> without settings.VISIBLE_TIMEZONES:
        with self.settings(VISIBLE_TIMEZONES=[]):
            with timezone.override(zoneinfo.ZoneInfo('Europe/Berlin')), translation.override(
                'de-de'
            ):
                html = human_timezone_datetime(dt=parse_dt('2000-01-01T00:00:00+0000'))
        self.assertHTMLEqual(html, '01.01.2000 01:00<br><small>(Europe/Berlin)</small>')



from django.core.management.base import BaseCommand
from django.urls import reverse

from bx_django_utils.admin_extra_views.registry import extra_view_registry


class Command(BaseCommand):
    """
    Manage command "admin_extra_views": Info about registered admin extra views
    """

    help = 'Information about registered admin extra views'

    def handle(self, *args, **options):
        view_count = 0
        for pseudo_app in sorted(extra_view_registry.pseudo_apps, key=lambda x: x.meta.name):
            self.stdout.write('_' * 100)
            self.stdout.write(f'Pseudo app: {pseudo_app.meta.name!r}')
            for view_class in sorted(pseudo_app.views, key=lambda x: x.meta.name):
                view_count += 1
                self.stdout.write(f' * Pseudo model: {view_class.meta.name!r}')
                url = reverse(view_class.meta.url_name)
                self.stdout.write(f' * url: {url!r}')
                self.stdout.write('')

        self.stdout.write(f'\nThere exists: {view_count} admin extra views.')



from django.core.cache.backends.base import MEMCACHE_MAX_KEY_LENGTH
from django.db import models
from django.utils.translation import gettext_lazy as _

from bx_django_utils.feature_flags.state import State
from bx_django_utils.feature_flags.utils import validate_cache_key
from bx_django_utils.models.timetracking import TimetrackingBaseModel


class FeatureFlagModel(TimetrackingBaseModel):
    """"""  # noqa - Don't add to README

    # Note: This model is only used as fallback on cache miss.
    #       It should not be directly used!

    cache_key = models.CharField(
        primary_key=True,
        max_length=MEMCACHE_MAX_KEY_LENGTH,
        verbose_name=_('FeatureFlagModel.cache_key.verbose_name'),
        help_text=_('FeatureFlagModel.cache_key.help_text'),
    )
    state = models.SmallIntegerField(
        choices=State.choices,
        verbose_name=_('FeatureFlagModel.state.verbose_name'),
        help_text=_('FeatureFlagModel.state.help_text'),
    )

    def full_clean(self, exclude=None, **kwargs):
        super().full_clean(exclude=exclude, **kwargs)

        if exclude is None or 'cache_key' not in exclude:
            validate_cache_key(cache_key=self.cache_key)

    def __str__(self):
        return f'{self.cache_key} {self.state}'

    class Meta:
        verbose_name = _('FeatureFlagModel.verbose_name')
        verbose_name_plural = _('FeatureFlagModel.verbose_name_plural')



from django import forms
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView

from bx_django_utils.admin_extra_views.base_view import AdminExtraViewMixin
from bx_django_utils.admin_extra_views.datatypes import AdminExtraMeta
from bx_django_utils.feature_flags.data_classes import FeatureFlag
from bx_django_utils.feature_flags.state import State


class ManageFeatureFlagsForm(forms.Form):
    """"""  # noqa - don't add in README

    cache_key = forms.ChoiceField(
        choices=((instance.cache_key, instance.human_name) for instance in FeatureFlag.values())
    )
    new_value = forms.ChoiceField(choices=State.choices)


class ManageFeatureFlagsBaseView(AdminExtraViewMixin, FormView):
    """
    Base admin extra view to manage all existing feature flags in admin.

    Intended to register this view yourself and set/check the user permissions
    according to your own requirements.
    """

    meta = AdminExtraMeta(name=_('Manage Feature Flags'), app_label='manage')
    template_name = 'admin_extra_views/manage_feature_flags.html'
    form_class = ManageFeatureFlagsForm

    def get_context_data(self, **context):
        feature_flags = []
        for instance in FeatureFlag.values():
            feature_flags.append(
                dict(
                    cache_key=instance.cache_key,
                    human_name=instance.human_name,
                    description=instance.description,
                    state=instance.state,
                    initial_state=instance.initial_state,
                    opposite_state=instance.opposite_state,
                )
            )
        if not feature_flags:
            messages.info(self.request, _('No feature flags currently registered'))
        context['feature_flags'] = feature_flags
        return super().get_context_data(**context)

    def form_valid(self, form):
        cache_key = form.cleaned_data['cache_key']
        new_value_str = form.cleaned_data['new_value']

        new_sate_value = int(new_value_str)
        new_state = State(new_sate_value)

        feature_flag = FeatureFlag.get_by_cache_key(cache_key)
        if feature_flag.opposite_state != new_state:
            messages.error(
                self.request,
                f'Current "{feature_flag.human_name}" state is already: {feature_flag.state.name}',
            )
        else:
            feature_flag.set_state(new_state)
            messages.success(
                self.request, f'Set "{feature_flag.human_name}" state to {feature_flag.state.name}'
            )
        return HttpResponseRedirect('.')



"""
    Example "Feature Flags"
"""
import inspect

from bx_django_utils.feature_flags.data_classes import FeatureFlag


foo_feature_flag = FeatureFlag(
    cache_key='foo',
    human_name='Foo',
    description=inspect.cleandoc(
        '''
        En-/disable example Feature flag `Foo`.
        Just for demonstrate.
        '''
    ),
    initial_enabled=True,
)

bar_feature_flag = FeatureFlag(
    cache_key='bar',
    human_name='Bar',
    description=inspect.cleandoc(
        '''
        En-/disable example Feature flag `Bar`.
        '''
    ),
    initial_enabled=False,
)



import sys
from pathlib import Path
from unittest import TestCase

from bx_py_utils.auto_doc import FnmatchExclude, assert_readme

import bx_django_utils


class Readme(TestCase):
    def test_auto_doc_in_readme(self):
        assert sys.version_info >= (3, 9), 'requires Python v3.9 or newer'
        root_path = Path(bx_django_utils.__file__).parent.parent
        readme_path = root_path / 'README.md'

        assert_readme(
            readme_path=readme_path,
            modules=['bx_django_utils'],
            exclude_func=FnmatchExclude('*/migrations/*'),
            start_marker_line='[comment]: <> (✂✂✂ auto generated start ✂✂✂)',
            end_marker_line='[comment]: <> (✂✂✂ auto generated end ✂✂✂)',
            start_level=2,
            link_template='https://github.com/boxine/bx_django_utils/blob/master/{path}#L{start}-L{end}',
        )



"""
    DEMO for bx_django_utils.view_utils.dynamic_menu_urls.DynamicViewMenu

    This DEMO menu is registeres in urls.py as root page.
    So see and use it, start dev. server and go to index page:

        bx_django_utils$ make start-dev-server
"""
from django.urls import reverse
from django.views.generic import RedirectView, TemplateView

from bx_django_utils.view_utils.dynamic_menu_urls import DynamicViewMenu


dynamic_view_menu = DynamicViewMenu()  # Init new menu instance


class MenuView(TemplateView):
    template_name = 'dynamic_view_menu/index.html'
    title = 'Dynamic View Menu - Index'

    def get_context_data(self):
        context = {
            'title': self.title,
            'menu': dynamic_view_menu.menu,
        }
        return super().get_context_data(**context)


class Redirect2AdminView(RedirectView):
    title = 'Go to Django Admin'

    def get_redirect_url(self, *args, **kwargs):
        return reverse('admin:index')


class DemoView1(TemplateView):
    template_name = 'dynamic_view_menu/demo.html'
    title = 'Dynamic View Menu - DEMO 1'

    def get_context_data(self):
        context = {
            'title': self.title,
            'content': 'DEMO 1 content',
        }
        return super().get_context_data(**context)


class DemoView2(TemplateView):
    template_name = 'dynamic_view_menu/demo.html'
    title = 'Dynamic View Menu - DEMO 2'

    def get_context_data(self):
        context = {
            'title': self.title,
            'content': 'DEMO 2 content',
        }
        return super().get_context_data(**context)


class DemoView3(TemplateView):
    template_name = 'dynamic_view_menu/demo.html'
    title = 'Dynamic View Menu - DEMO 3'

    def get_context_data(self):
        context = {
            'title': self.title,
            'content': 'DEMO 3 content',
        }
        return super().get_context_data(**context)


# Add views to menu and urls:
dynamic_view_menu.add_views(
    app_name='test_app',  # same as: test_app.urls.app_name !
    menu=(
        ('Section 1', {
            'views': (
                # (<ViewClass>, <url-name>)
                (Redirect2AdminView, 'redirect2admin'),  # url name -> 'test_app:redirect2admin'
            )
        }),
        ('Section 2', {
            'views': (
                (DemoView1, 'demo1'),
                (DemoView2, 'demo2'),
            )
        }),
        ('Section 3', {
            'views': (
                (DemoView3, 'demo3'),
            )
        }),
    )
)



from bx_django_utils.admin_utils.filters import NotAllSimpleListFilter


class NotAllSimpleListFilterDemo(NotAllSimpleListFilter):
    title = 'blank_field set'
    parameter_name = 'blank_field'

    def lookups(self, request, model_admin):
        return (
            ('all', 'All'),
            (None, 'Yes'),  # None -> default filter!
            ('no', 'No'),
        )

    def queryset(self, request, queryset):
        if self.value() == 'all':  # -> don't filter
            return queryset
        elif self.value() == 'no':  # -> Only without a "blank_field" value
            return queryset.filter(blank_field='')

        # Default -> Only items with a "blank_field" value
        return queryset.exclude(blank_field='')



import json
from functools import partial
from typing import Union

from django import forms
from django.conf import settings
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.db import models
from django.forms.fields import InvalidJSONInput
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _

from bx_django_utils.models.manipulate import CreateOrUpdateResult, FieldUpdate, update_model_field


class TranslationWidget(forms.Widget):
    template_name = 'bx_django_utils/translation_input.html'

    def __init__(self, language_codes: tuple, attrs=None):
        self.language_codes = language_codes
        super().__init__(attrs=attrs)

    def format_value(self, value):
        # for some reason, value is always a string. convert back to dict.
        return json.loads(value)

    def value_from_datadict(self, data, files, name):
        fieldnames = {f'{name}__{code}': code for code in self.language_codes}  # e.g. "fieldname__de-de" -> "de-de"
        return {  # e.g. "de-de" -> "the translation string"
            fieldnames[fieldname]: data[fieldname] for fieldname in data if fieldname in fieldnames
        }

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['widget']['language_codes'] = self.language_codes
        return context


class MultilineTranslationWidget(TranslationWidget):
    template_name = 'bx_django_utils/translation_input_multiline.html'


class TranslationFormField(forms.JSONField):
    """
    Default form field for TranslationField.
    """

    widget = None  # set by __init__

    def __init__(self, *args, language_codes, widget_class=TranslationWidget, **kwargs):
        self.widget = widget_class
        # self.widget can be a class or an instance.
        # abuse that to pass the language codes of the related TranslationField
        # to the widget.
        self.widget = self.widget(language_codes=language_codes)
        super().__init__(*args, **kwargs)

    def bound_data(self, data, initial):
        if self.disabled:
            return initial
        if data is None:
            return None
        if isinstance(data, str):
            try:
                return json.loads(data, cls=self.decoder)
            except json.JSONDecodeError:
                return InvalidJSONInput(data)
        return data


class FieldTranslation(dict):
    """
    Dict-like container that maps language codes to a translated string.
    Is returned when accessing the value of TranslationField.
    """

    UNTRANSLATED = '<UNTRANSLATED>'

    def get_first(self, codes):
        """
        Returns the first translation found for the given language codes.
        """
        for code in codes:
            if translation := self.get(code):
                return translation
        return self.UNTRANSLATED

    def __repr__(self):
        return f'FieldTranslation({super().__repr__()})'


class TranslationField(models.JSONField):
    """
    A field designed to hold translations for a given set of language codes.
    You may optionally define which widget this field should render as by default via
    the `widget_class` kwarg. The widget must always return a dict-like value, so
    it can be stored as JSON in the database.

    TranslationFields always have null=True. If no value is set, an empty dict will
    be used instead. This ensures a consistent way to access the data.

    Accessing this field's value returns a FieldTranslation instance instead of
    a regular dict (much like FileField returns a FieldFile instance).
    """

    default_error_messages = {
        # Replace "This field cannot be blank." a more appropriate message:
        'blank': _('At least one translation is required.'),
    }

    def __init__(self, language_codes: tuple, *args, **kwargs):
        kwargs['null'] = False
        kwargs['default'] = FieldTranslation
        self.language_codes = language_codes
        self.widget_class = kwargs.pop('widget_class', TranslationWidget)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        for override in ('null', 'default'):
            # TranslationField hardcodes values for these kwargs in __init__,
            # so omit them from deconstruction
            kwargs.pop(override, None)
        args = [self.language_codes, *args]
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        value = super().from_db_value(value, expression, connection)
        if value is None:
            return FieldTranslation()
        value = remove_empty_translations(value)  # Ignore empty translation from DB
        return FieldTranslation(value)

    def get_db_prep_save(self, value, *args, **kwargs):
        value = remove_empty_translations(value)  # Don't store empty translations to DB
        return super().get_db_prep_save(value, *args, **kwargs)

    def clean(self, value, model_instance):
        value = super().clean(value, model_instance)
        value = remove_empty_translations(value)
        existing_codes = value.keys()
        unknown_codes = existing_codes - self.language_codes
        if unknown_codes:
            raise ValidationError(f'Unknown translation language(s): {", ".join(sorted(unknown_codes))}')
        return value

    def to_python(self, value):
        if not value:
            return FieldTranslation()

        if isinstance(value, str):
            try:
                value = json.loads(value, cls=self.decoder)
            except json.JSONDecodeError as err:
                raise ValidationError(f'Invalid JSON data: {err}')
        if not isinstance(value, dict):
            raise ValidationError(f'Invalid input type: {type(value).__name__}')

        value = remove_empty_translations(value)
        return FieldTranslation(value)

    def formfield(self, **kwargs):
        kwargs['form_class'] = TranslationFormField
        kwargs['widget_class'] = self.widget_class
        kwargs['language_codes'] = self.language_codes
        return super().formfield(**kwargs)


class TranslationFieldAdmin(admin.ModelAdmin):
    """
    Provides drop-in support for ModelAdmin classes that want to display TranslationFields
    in their changelist.

    Since TranslationFields are essentially a JSONField, the Django admin will _always_ render
    them as a serialized JSON objects (see django.contrib.admin.utils.display_for_field).
    This is not helpful in the context of TranslationField, so this base class does some
    patching to display one of the translations instead.
    """

    @staticmethod
    def get_translation_order(fallback_codes):
        """
        Determines the order in which a suitable translation is picked from the value of a TranslationField.
        Can be overriden by inheriting admin classes, e.g. when using language codes other
        than those the Django translation system uses.
        """
        return [
            get_language(),  # equal to current locale of user
            settings.LANGUAGE_CODE,
            *fallback_codes,
        ]

    def _get_translated_field_display(self, obj, fieldname, field_codes):
        field = getattr(obj, fieldname)
        order = self.get_translation_order(field_codes)
        trans = field.get_first(order)
        return trans

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # collect names of fields that are TranslationFields.
        # also save the translation codes for each field.
        # these will be used below for resolving a user-appropriate translation.
        translation_fields = []
        translation_codes = {}
        for field in self.model._meta.fields:
            if not isinstance(field, TranslationField):
                continue
            translation_fields.append(field.name)
            translation_codes[field.name] = field.language_codes

        # add a 'get_<fieldname>' method to the admin for each TranslationField found in the last step.
        # the method returns its most suitable translation.
        for field in translation_fields:
            resolve_func = partial(
                self._get_translated_field_display,
                fieldname=field,
                field_codes=translation_codes[field],
            )
            resolve_func.__name__ = field
            setattr(self, f'get_{field}', resolve_func)

        def patch_fieldnames(fieldnames):
            return [f'get_{field}' if field in translation_fields else field for field in fieldnames]

        # patch each list of fieldnames considered by the changelist to use the new 'get_<fieldname>' methods.
        self.list_display = patch_fieldnames(self.list_display)
        self.list_display_links = patch_fieldnames(self.list_display_links)
        self.readonly_fields = patch_fieldnames(self.readonly_fields)


def create_or_update_translation_callback(*, instance, field_name, old_value, new_value, result: CreateOrUpdateResult):
    """
    Callback for create_or_update2() for TranslationField, that will never remove existing translation.
    """
    if old_value == new_value:
        # Nothing to update -> we are done.
        return

    model_field = instance._meta.get_field(field_name)
    if not isinstance(model_field, TranslationField):
        # Normal model field -> use origin callback
        return update_model_field(
            instance=instance,
            field_name=field_name,
            old_value=old_value,
            new_value=new_value,
            result=result,
        )

    # Merge translations:
    merged_value = merge_translations(old_value, new_value)

    if old_value == merged_value:
        # Nothing to update -> we are done.
        return

    # The model field value has been changed -> set it on the model instance:
    setattr(instance, field_name, merged_value)

    # Expand the list of updated fields, used for result and save() call:
    result.updated_fields.append(field_name)

    # Store update information:
    result.update_info.append(FieldUpdate(field_name=field_name, old_value=old_value, new_value=merged_value))


def remove_empty_translations(translations: Union[dict, FieldTranslation]) -> FieldTranslation:
    """
    Remove all empty/None from a FieldTranslation, e.g.:

    >>> remove_empty_translations({'de-de': 'Hallo', 'en': 'Hello', 'es':'', 'mx': None})
    FieldTranslation({'de-de': 'Hallo', 'en': 'Hello'})
    """
    filtered = {lang_code: text for lang_code, text in translations.items() if text}
    return FieldTranslation(filtered)


def merge_translations(
    translations1: Union[dict, FieldTranslation], translations2: Union[dict, FieldTranslation]
) -> FieldTranslation:
    """
    Merge two FieldTranslation and ignore all empty/None values, e.g.:

    >>> merge_translations({'de': 'Hallo', 'en': '', 'es': 'Hola'}, {'de': '', 'es': 'HOLA'})
    FieldTranslation({'de': 'Hallo', 'es': 'HOLA'})
    """
    merged = remove_empty_translations(translations1)
    merged.update(remove_empty_translations(translations2))
    return FieldTranslation(merged)



from http.cookies import SimpleCookie

from django.conf import settings
from django.test import TestCase
from playwright.sync_api import BrowserContext, expect

from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin
from bx_django_utils.test_utils.playwright import PlaywrightTestCase
from bx_django_utils.test_utils.users import make_minimal_test_user


class UserTimezoneTestCase(HtmlAssertionMixin, TestCase):
    """
    Integration tests for user timezone into test project
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.staff_user = make_minimal_test_user(username='foobar', is_staff=True, permissions=())

    def test_admin_index(self):
        self.client.cookies = SimpleCookie({'UserTimeZone': 'Europe/London'})
        self.client.force_login(self.staff_user)

        response = self.client.get('/admin/')
        self.assert_html_parts(
            response,
            parts=(
                '<title>Site administration | Django site admin</title>',
                '<span>Active Timezone:&nbsp;Europe/London</span>',
            ),
        )


class UserTimezonePlaywrightMixin:
    """
    Run User Timezone tests with Firefox and Chromium, via:
        * UserTimezoneFirefoxTestCase
        * UserTimezoneChromiumTestCase
    """

    def test_user_timezone(self):
        context: BrowserContext = self.browser.new_context(
            ignore_https_errors=True,  # StaticLiveServerTestCase has no valid certificate ;)
            locale='en_US',
            timezone_id='Europe/Berlin',
        )
        with context.new_page() as page:
            # Use our "Dynamic View Menu" index page from our "test_app":
            page.goto(f'{self.live_server_url}')
            expect(page).to_have_title('Dynamic View Menu - Index | Django site admin')

            # UserTimeZone cookie set?
            cookies = context.cookies()
            self.assertEqual(
                cookies,
                [
                    {
                        'domain': 'localhost',
                        'expires': -1,
                        'httpOnly': False,
                        'name': 'UserTimeZone',
                        'path': '/',
                        'sameSite': 'Lax',
                        'secure': False,
                        'value': 'Europe/Berlin',
                    }
                ],
            )

            # First request: The cookie is set in the client but not yet sent to the server.
            # So we get the default timezone:
            assert settings.TIME_ZONE == 'UTC'  # Check the default timezone
            expect(page.locator('text=Active Timezone: UTC')).to_be_visible()

            # Just reload the page to send the cookie to the server:
            page.reload()

            # Now the timezone should be active:
            expect(page.locator('text=Active Timezone: Europe/Berlin')).to_be_visible()


class UserTimezoneFirefoxTestCase(UserTimezonePlaywrightMixin, PlaywrightTestCase):
    BROWSER_NAME = 'firefox'


class UserTimezoneChromiumTestCase(UserTimezonePlaywrightMixin, PlaywrightTestCase):
    BROWSER_NAME = 'chromium'



from django.urls import path

from bx_django_utils_tests.test_app.views.dynamic_view_menu import MenuView, dynamic_view_menu


app_name = 'test_app'


urlpatterns = [
    path('', MenuView.as_view(), name='index'),
] + dynamic_view_menu.get_urls()  # Add all Fixup views



from django.core.cache import cache
from django.test import SimpleTestCase

from bx_django_utils.test_utils.cache import ClearCacheMixin, MockCache


class MockCacheTestCase(SimpleTestCase):
    def test_mock_cache(self):
        with MockCache() as cache_mock:
            cache.set(key='foo', value=1, timeout=123)
            cache.set(key='bar', value=2, timeout=123)
            assert cache.get(key='foo') == 1
            assert cache.get_many(keys=('foo', 'bar'), version=1) == {'foo': 1, 'bar': 2}
            assert cache_mock.raw_data == {':1:bar': 2, ':1:foo': 1}
            assert cache_mock.data == {'foo': 1, 'bar': 2}
            cache.delete(key='foo')
        assert cache_mock.data == {'bar': 2}

        with MockCache() as cache_mock:
            cache.add(key='test-add', value=1)
            cache.add(key='test-add', value=999)
        assert cache_mock.data == {'test-add': 1}

        with MockCache() as cache_mock:
            cache.set(key='test-incr', value=10)
            cache.incr(key='test-incr')  # 10+1=11
            cache.incr(key='test-incr')  # 11+1=12
            cache.decr(key='test-incr')  # 12-1=11
        assert cache_mock.raw_data == {':1:test-incr': 11}
        assert cache_mock.data == {'test-incr': 11}

        with MockCache() as cache_mock:
            assert cache.get_or_set(key='test-get-or-set', default=1) == 1
            assert cache.get_or_set(key='test-get-or-set', default=999) == 1
        assert cache_mock.raw_data == {':1:test-get-or-set': 1}
        assert cache_mock.data == {'test-get-or-set': 1}

        with MockCache() as cache_mock:
            cache.set(key='version-test', value='a', version=1)
            cache.set(key='version-test', value='b', version=2)
        assert cache_mock.raw_data == {':1:version-test': 'a', ':2:version-test': 'b'}
        assert cache_mock.data == {'version-test': 'b'}


class ClearCacheTestCase(ClearCacheMixin, SimpleTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cache.set(key='foo', value=1, timeout=0)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cache.set(key='foo', value=2, timeout=0)

    def setUp(self):
        super().setUp()
        assert cache.get(key='foo') is None

    def tearDown(self):
        super().tearDown()
        assert cache.get(key='foo') is None

    def test_check1(self):
        assert cache.get(key='foo') is None
        cache.set(key='foo', value=3, timeout=0)

    def test_check2(self):
        assert cache.get(key='foo') is None
        cache.set(key='foo', value=4, timeout=0)



import logging
import zoneinfo

from django.utils import timezone


logger = logging.getLogger(__name__)


class UserTimezoneMiddleware:
    """
    Activate Timezone by "UserTimeZone" cookie
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if raw_time_zone := request.COOKIES.get('UserTimeZone'):
            try:
                tzinfo = zoneinfo.ZoneInfo(raw_time_zone)
            except zoneinfo.ZoneInfoNotFoundError:
                logger.error('Unknown user time zone %r', raw_time_zone)
            else:
                # Response with user timezone:
                with timezone.override(tzinfo):
                    response = self.get_response(request)
                return response

        # Response with default timezone:
        return self.get_response(request)



import os

from django.contrib.auth import get_user_model
from django.core.management import BaseCommand, call_command


class Command(BaseCommand):
    """
    Expand django.contrib.staticfiles runserver
    """
    help = "Setup test project and run django developer server"

    def verbose_call(self, command, *args, **kwargs):
        self.stderr.write("_" * 79)
        self.stdout.write(f"Call {command!r} with: {args!r} {kwargs!r}")
        call_command(command, *args, **kwargs)

    def handle(self, *args, **options):

        if "RUN_MAIN" not in os.environ:
            # RUN_MAIN added by auto reloader, see: django/utils/autoreload.py

            # Create migrations for our test app
            # But these migrations should never commit!
            # On changes: Just delete the SQLite file and start fresh ;)
            self.verbose_call("makemigrations")

            self.verbose_call("migrate")

            # django.contrib.staticfiles.management.commands.collectstatic.Command
            self.verbose_call("collectstatic", interactive=False, link=True)

            User = get_user_model()
            qs = User.objects.filter(is_active=True, is_superuser=True)
            if qs.count() == 0:
                self.verbose_call("createsuperuser")

        self.verbose_call("runserver", use_threading=False, use_reloader=True, verbosity=2)



import datetime
from unittest import mock

from django.test import SimpleTestCase
from django.utils import timezone

from bx_django_utils.test_utils.datetime import MockDatetimeGenerator


class MockDatetimeGeneratorTestCase(SimpleTestCase):
    def test_daetime_timezone_now_mock(self):
        with mock.patch.object(timezone, 'now', MockDatetimeGenerator()):
            assert timezone.now() == datetime.datetime(
                2001, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
            )
            assert timezone.now() == datetime.datetime(
                2002, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
            )
            assert timezone.now() == datetime.datetime(
                2003, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
            )

        offset = datetime.timedelta(minutes=1)
        with mock.patch.object(timezone, 'now', MockDatetimeGenerator(offset=offset)):
            assert timezone.now() == datetime.datetime(
                2000, 1, 1, 0, 1, tzinfo=datetime.timezone.utc
            )
            assert timezone.now() == datetime.datetime(
                2000, 1, 1, 0, 2, tzinfo=datetime.timezone.utc
            )
            assert timezone.now() == datetime.datetime(
                2000, 1, 1, 0, 3, tzinfo=datetime.timezone.utc
            )



from copy import deepcopy
from typing import Type

from django.db.models import Model, Q, QuerySet
from django.db.models.sql.where import NothingNode


def remove_filter(queryset: QuerySet, lookup: str) -> QuerySet:
    """
    Remove an applied .filter() from a QuerySet
    """
    if queryset.query.is_empty():
        # Nothing to remove if queryset is empty e.g.: ...objects.none()
        return queryset

    queryset = deepcopy(queryset)  # remove the QuerySet's cache

    query = queryset.query

    clause, _ = query._add_q(Q(**{lookup: None}), query.used_aliases)

    def filter_lookups(node):
        if hasattr(node, 'lhs'):
            return node.lhs.target != clause.children[0].lhs.target

        if isinstance(node, NothingNode):
            return False

        return len(list(filter(filter_lookups, node.children))) == len(node.children)

    query.where.children = list(filter(filter_lookups, query.where.children))

    return queryset


def remove_model_filter(queryset: QuerySet, model: Type[Model]) -> QuerySet:
    """
    Remove an applied .filter() from a QuerySet if it contains references to the specified model
    """
    if queryset.query.is_empty():
        # Nothing to remove if queryset is empty e.g.: ...objects.none()
        return queryset

    queryset = deepcopy(queryset)  # remove the QuerySet's cache
    query = queryset.query

    def filter_lookups(node):
        if hasattr(node, 'lhs'):
            return node.lhs.target.model != model

        if isinstance(node, NothingNode):
            return False

        return len(list(filter(filter_lookups, node.children))) == len(node.children)

    query.where.children = list(filter(filter_lookups, query.where.children))

    return queryset



import json
import re
from itertools import product

from bx_py_utils.test_utils.snapshot import assert_html_snapshot
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.test import TestCase
from django.utils import translation

from bx_django_utils.models.manipulate import FieldUpdate, create_or_update2
from bx_django_utils.test_utils.users import make_test_user
from bx_django_utils.translation import (
    FieldTranslation,
    TranslationField,
    TranslationFormField,
    TranslationWidget,
    create_or_update_translation_callback,
)
from bx_django_utils_tests.test_app.models import RawTranslatedModel, TranslatedModel


class TranslationFieldTestCase(TestCase):
    def test_stores_only_wanted_languages(self):
        obj = TranslatedModel(
            translated={'de-de': 'Hallo', 'en-us': 'Hello', 'es': 'Hola', 'it': 'Ciao'},
            translated_multiline={'de-de': 'Hallo'},
        )
        with self.assertRaisesMessage(ValidationError, 'Unknown translation language(s): it'):
            obj.full_clean()

        obj = TranslatedModel(
            translated={'de-de': 'Hallo'},
            translated_multiline={'de-de': 'Hallo', 'en-us': 'Hello', 'es': 'Hola', 'it-it': 'Ciao', 'xx': 'foo'},
        )
        with self.assertRaisesMessage(ValidationError, 'Unknown translation language(s): it-it, xx'):
            obj.full_clean()

    def test_field_translation(self):
        translated = FieldTranslation()
        self.assertEqual(translated.get_first([1, 2, 3]), FieldTranslation.UNTRANSLATED)
        self.assertEqual(repr(translated), 'FieldTranslation({})')

        translated = FieldTranslation({'de-de': 'Hallo'})
        self.assertEqual(repr(translated), "FieldTranslation({'de-de': 'Hallo'})")
        self.assertEqual(translated.get_first(['foo', 'bar', 'en-gb', 'de-de']), 'Hallo')

        obj1 = TranslatedModel.objects.create(translated={'de-de': 'Hallo 1', 'en-us': 'Hello 1'})
        obj1.full_clean()
        self.assertEqual(repr(obj1.translated), "FieldTranslation({'de-de': 'Hallo 1', 'en-us': 'Hello 1'})")

        obj2 = TranslatedModel.objects.create(translated={'de-de': 'Hallo 2', 'en-us': 'Hello 2'})
        obj2.full_clean()

        self.assertNotEqual(obj1.translated, obj2.translated)

        qs = TranslatedModel.objects.order_by('translated').values('translated')
        self.assertQuerysetEqual(
            qs,
            [
                {'translated': FieldTranslation({'de-de': 'Hallo 1', 'en-us': 'Hello 1'})},
                {'translated': FieldTranslation({'de-de': 'Hallo 2', 'en-us': 'Hello 2'})},
            ],
        )

    def test_blank(self):
        obj = TranslatedModel()
        with self.assertRaisesMessage(ValidationError, 'At least one translation is required.'):
            obj.full_clean()

        obj.translated = {}
        with self.assertRaisesMessage(ValidationError, 'At least one translation is required.'):
            obj.full_clean()

        obj.translated = {'de-de': ''}
        with self.assertRaisesMessage(ValidationError, 'At least one translation is required.'):
            obj.full_clean()

        obj.translated = {'de-de': 'a value'}
        obj.full_clean()

    def test_empty_translations_from_db(self):
        # Store empty values using the RawTranslatedModel:
        RawTranslatedModel.objects.create(translated={'de-de': 'A value', 'en-gb': '', 'en-us': None})
        raw_obj = RawTranslatedModel.objects.first()
        self.assertEqual(raw_obj.translated, {'de-de': 'A value', 'en-gb': '', 'en-us': None})

        # Empty values from DB will be ignored:
        obj = TranslatedModel.objects.first()
        self.assertEqual(obj.translated, FieldTranslation({'de-de': 'A value'}))

    def test_empty_translations_to_db(self):
        # Clean removes double entries:
        obj = TranslatedModel(translated={'de-de': 'A value', 'en-gb': '', 'en-us': None})
        obj.full_clean()
        self.assertEqual(obj.translated, {'de-de': 'A value'})

        # It's also not possible to store empty translation in this way:
        TranslatedModel.objects.create(translated={'de-de': 'A value', 'en-gb': '', 'en-us': None})
        raw_obj = RawTranslatedModel.objects.first()
        self.assertEqual(raw_obj.translated, {'de-de': 'A value'})

    def test_create_or_update(self):
        # Create via create_or_update2():
        result = create_or_update2(ModelClass=TranslatedModel, translated={'de-de': 'Hallo'})
        self.assertTrue(result.created)
        instance = result.instance
        self.assertEqual(instance.translated, {'de-de': 'Hallo'})
        self.assertEqual(instance.not_translated, 'A Default Value')

        # Update existing entry:
        result = create_or_update2(
            ModelClass=TranslatedModel,
            lookup=dict(pk=instance.pk),
            translated={'de-de': 'Hallo', 'en-us': 'Hello'},
            not_translated='Not the default.',
        )
        self.assertFalse(result.created)  # updated?
        instance = result.instance
        self.assertEqual(result.updated_fields, ['translated', 'not_translated'])
        self.assertEqual(
            result.update_info,
            [
                FieldUpdate(
                    field_name='translated',
                    old_value=FieldTranslation({'de-de': 'Hallo'}),
                    new_value={'de-de': 'Hallo', 'en-us': 'Hello'},
                ),
                FieldUpdate(field_name='not_translated', old_value='A Default Value', new_value='Not the default.'),
            ],
        )
        self.assertEqual(instance.translated, {'de-de': 'Hallo', 'en-us': 'Hello'})
        self.assertEqual(instance.not_translated, 'Not the default.')

        # We have a special callback to avoid deleting translations:
        result = create_or_update2(
            ModelClass=TranslatedModel,
            lookup=dict(pk=instance.pk),
            update_model_field_callback=create_or_update_translation_callback,
            translated={'de-de': 'Hallo !', 'es': 'Hola'},
        )
        self.assertFalse(result.created)  # updated?
        instance = result.instance
        self.assertEqual(result.updated_fields, ['translated'])
        self.assertEqual(
            result.update_info,
            [
                FieldUpdate(
                    field_name='translated',
                    old_value=FieldTranslation({'de-de': 'Hallo', 'en-us': 'Hello'}),
                    new_value={'de-de': 'Hallo !', 'en-us': 'Hello', 'es': 'Hola'},
                )
            ],
        )
        self.assertEqual(
            instance.translated,
            FieldTranslation(
                {
                    'de-de': 'Hallo !',  # <<< changed
                    'en-us': 'Hello',  # <<< untouched
                    'es': 'Hola',  # <<< added
                }
            ),
        )
        self.assertEqual(instance.not_translated, 'Not the default.')  # unchanged?

    def test_language_codes_order(self):
        language_codes = ('fr-fr', 'de-de', 'en-us')
        model_field = TranslationField(language_codes=language_codes)
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, TranslationFormField)

        widget = form_field.widget
        self.assertIsInstance(widget, TranslationWidget)

        # language_codes pass to the widget?
        self.assertIsInstance(widget.language_codes, tuple)
        self.assertIs(widget.language_codes, language_codes)

        html = widget.render(name='foobar', value=b'{}')

        # Is the order the same as specifies in TranslationField() ?
        self.assertEqual(tuple(re.findall(r'<td>(.{5})</td>', html)), language_codes)
        assert_html_snapshot(got=html, validate=False)

    def test_query(self):
        TranslatedModel.objects.create(translated={'de-de': 'Hallo 1'})
        TranslatedModel.objects.create(translated={'de-de': 'Hallo 2', 'en-us': 'Hello 2'})
        TranslatedModel.objects.create(translated={'de-de': 'Hallo 3', 'en-us': 'Hello 3'})
        self.assertEqual(
            list(TranslatedModel.objects.filter(translated={'de-de': 'Hallo 1'}).values('translated')),
            [{'translated': FieldTranslation({'de-de': 'Hallo 1'})}],
        )

        qs = TranslatedModel.objects.filter(**{'translated__de-de': 'Hallo 1'})
        self.assertEqual(
            list(qs.values('translated')),
            [{'translated': FieldTranslation({'de-de': 'Hallo 1'})}],
        )

        qs = TranslatedModel.objects.order_by('translated').filter(
            Q(**{'translated__de-de': 'Hallo 1'}) | Q(**{'translated__en-us': 'Hello 3'})
        )
        self.assertEqual(
            list(qs.values('translated')),
            [
                {'translated': FieldTranslation({'de-de': 'Hallo 1'})},
                {'translated': FieldTranslation({'de-de': 'Hallo 3', 'en-us': 'Hello 3'})},
            ],
        )


class TranslationAdminTestCase(TestCase):
    # test admin base class and widgets

    def create_test_obj(self):
        obj = TranslatedModel(
            translated={'de-de': 'Zug', 'en-us': 'Train', 'es': 'Tren'},
            translated_multiline={
                'de-de': 'Ein Zug ist sehr schnell.\nZüge können schneller sein als Autos.',
                'en-us': 'A train is very fast.\nTrains can be faster than cars.',
                'es': 'Un tren es muy rápido.\nLos trenes pueden ser más rápidos que los automóviles.',
            },
        )
        obj.full_clean()
        obj.save()
        return obj

    def setUp(self):
        super().setUp()
        self.user = make_test_user(is_superuser=True)
        self.client.force_login(self.user)

    def test_changelist(self):
        obj = self.create_test_obj()

        CODE = 'de-de'
        with translation.override(CODE):
            response = self.client.get('/admin/test_app/translatedmodel/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'], 'text/html; charset=utf-8')
        response_content = response.content.decode()
        self.assertIn('class="field-get_translated"', response_content)
        self.assertInHTML(obj.translated[CODE], response_content)
        self.assertIn('class="field-get_translated_multiline"', response_content)
        self.assertInHTML(obj.translated_multiline[CODE], response_content)

    def test_change(self):
        obj = self.create_test_obj()

        # view existing record's change form
        response = self.client.get(f'/admin/test_app/translatedmodel/{obj.pk}/change/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'], 'text/html; charset=utf-8')
        response_content = response.content.decode()

        fieldname = 'translated'
        for code in ['de-de', 'en-us', 'es']:
            self.assertInHTML(
                (
                    f'<input'
                    f' id="id_{fieldname}__{code}"'
                    f' type="text"'
                    f' name="{fieldname}__{code}"'
                    f' value="{getattr(obj, fieldname)[code]}">'
                ),
                response_content,
            )
        fieldname = 'translated_multiline'
        for code in ['de-de', 'en-us', 'es']:
            self.assertInHTML(
                (
                    f'<textarea'
                    f' id="id_{fieldname}__{code}"'
                    f' name="{fieldname}__{code}">'
                    f'{getattr(obj, fieldname)[code]}'
                    f'</textarea>'
                ),
                response_content,
            )

        # change all translations via form
        data = {
            'initial-translated': json.dumps(obj.translated),
            'initial-translated_multiline': json.dumps(obj.translated_multiline),
            'not_translated': 'Foo Bar',
        }
        for code in ['de-de', 'en-us', 'es']:
            data[f'translated__{code}'] = obj.translated[code].upper()
            data[f'translated_multiline__{code}'] = obj.translated_multiline[code].upper()
        response = self.client.post(f'/admin/test_app/translatedmodel/{obj.pk}/change/', data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'], 'text/html; charset=utf-8')
        messages = response.context['messages']
        self.assertEqual(len(messages), 1)
        self.assertIn('was changed successfully', [str(m) for m in messages][0])

        obj.refresh_from_db()
        self.assertEqual(obj.translated, {'de-de': 'ZUG', 'en-us': 'TRAIN', 'es': 'TREN'})
        self.assertEqual(
            obj.translated_multiline,
            {
                'de-de': 'EIN ZUG IST SEHR SCHNELL.\nZÜGE KÖNNEN SCHNELLER SEIN ALS AUTOS.',
                'en-us': 'A TRAIN IS VERY FAST.\nTRAINS CAN BE FASTER THAN CARS.',
                'es': 'UN TREN ES MUY RÁPIDO.\nLOS TRENES PUEDEN SER MÁS RÁPIDOS QUE LOS AUTOMÓVILES.',
            },
        )

    def test_add(self):
        # attempt to save empty form
        response = self.client.post('/admin/test_app/translatedmodel/add/', follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'], 'text/html; charset=utf-8')
        self.assertInHTML(
            '<li>This field is required.</li>',
            response.content.decode(),
            count=2,  # error from blank translated and not_translated TranslatedModel fields
        )
        self.assertEqual(TranslatedModel.objects.count(), 0)

        # save with some data
        data = {
            'initial-translated': '{}',
            'initial-translated_multiline': '{}',
            'not_translated': 'Foo Bar',
            '_continue': 'Save and continue editing',
        }
        for code, field in product(['de-de', 'en-us', 'es'], ['translated', 'translated_multiline']):
            expr = f'{field}__{code}'
            data[expr] = expr
        response = self.client.post('/admin/test_app/translatedmodel/add/', data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'], 'text/html; charset=utf-8')
        messages = response.context['messages']
        self.assertEqual(len(messages), 1)
        self.assertIn('was added successfully', [str(m) for m in messages][0])

        # was persisted?
        self.assertEqual(TranslatedModel.objects.count(), 1)
        obj = TranslatedModel.objects.first()
        self.assertEqual(
            obj.translated,
            {
                'de-de': 'translated__de-de',
                'en-us': 'translated__en-us',
                'es': 'translated__es',
            },
        )
        self.assertEqual(
            obj.translated_multiline,
            {
                'de-de': 'translated_multiline__de-de',
                'en-us': 'translated_multiline__en-us',
                'es': 'translated_multiline__es',
            },
        )
        self.assertEqual(obj.not_translated, 'Foo Bar')



import datetime

from django.conf import settings
from django.template.loader import render_to_string


def human_timezone_datetime(dt):
    """
    Render a datetime with timezone information.
    """
    if not dt:
        return '-'
    assert isinstance(dt, datetime.datetime)
    context = {
        'dt': dt,
        'timezones': settings.VISIBLE_TIMEZONES,
    }
    return render_to_string(
        template_name='user_timezone/human_timezone_datetime.html',
        context=context,
    )



from django.core.checks import Error, Warning, register
from django.urls import NoReverseMatch, reverse

from bx_django_utils.admin_extra_views.registry import extra_view_registry
from bx_django_utils.admin_extra_views.site import ExtraViewAdminSite


class _ByPassConditions:
    def __enter__(self):
        extra_view_registry._by_pass_conditions = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        extra_view_registry._by_pass_conditions = False


@register()
def admin_extra_views_check(app_configs, **kwargs):
    errors = []

    view_count = 0
    for view_class in extra_view_registry:
        try:
            url = reverse(view_class.meta.url_name)
        except NoReverseMatch as err:
            errors.append(
                Error(
                    msg=f'Admin extra views URL reverse error with {view_class.__name__!r}',
                    hint='Have you include "extra_view_registry.get_urls()" to your urls.py ?',
                    obj=err,
                    id='admin_extra_views.E001',
                )
            )
        else:
            assert url
            view_count += 1

    if view_count == 0:
        errors.append(
            Warning(
                msg='No admin extra views registered!',
                id='admin_extra_views.W001',
            )
        )

    for pseudo_app in extra_view_registry.pseudo_apps:
        admin_site = pseudo_app.admin_site
        if not isinstance(admin_site, ExtraViewAdminSite):
            errors.append(
                Error(
                    msg=(
                        f'Pseudo app {pseudo_app.meta.name!r} error:'
                        f' Admin site {admin_site.__class__.__name__!r}'
                        ' is not a instance of ExtraViewAdminSite!'
                    ),
                    hint='Use CustomAdminConfig or add "site" to PseudoApp definition',
                    id='admin_extra_views.E002',
                )
            )

    return errors



import os
import unittest
from pathlib import Path

import bx_django_utils


# Hacky way to display more "assert"-Context in failing tests:
unittest.util._MAX_LENGTH = os.environ.get('UNITTEST_MAX_LENGTH', 200)


PACKAGE_ROOT = Path(bx_django_utils.__file__).parent.parent



from django.conf import settings
from django.test import SimpleTestCase, override_settings

from bx_django_utils.user_timezone.checks import user_timezone_check


class ChecksTestCase(SimpleTestCase):
    def test_w001(self):
        errors = user_timezone_check(app_configs=None)
        self.assertEqual(errors, [])

        with self.settings(MIDDLEWARE=[]):
            errors = user_timezone_check(app_configs=None)
        self.assertEqual(len(errors), 1)
        warning = errors[0]
        self.assertEqual(warning.id, 'user_timezone.W001')
        self.assertEqual(warning.msg, 'Missing UserTimezoneMiddleware')

    @override_settings()
    def test_e001(self):
        del settings.VISIBLE_TIMEZONES

        errors = user_timezone_check(app_configs=None)
        self.assertEqual(len(errors), 1)
        error = errors[0]
        self.assertEqual(error.id, 'user_timezone.E001')
        self.assertEqual(error.msg, 'Your settings has no "VISIBLE_TIMEZONES" defined!')

    @override_settings(VISIBLE_TIMEZONES=['Europe/Berlin', 'non/sense'])
    def test_e002(self):
        errors = user_timezone_check(app_configs=None)
        self.assertEqual(len(errors), 1)
        error = errors[0]
        self.assertEqual(error.id, 'user_timezone.E002')
        self.assertEqual(error.msg, "settings.VISIBLE_TIMEZONES entry: 'non/sense' is invalid!")



import datetime
import json
import time

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.encoding import force_str

from bx_django_utils.stacktrace import get_stacktrace


try:
    from psycopg2._json import Json as PostgresJson
except ImportError:
    PostgresJson = None


class RecordingCursorWrapper:
    """
    An implementation of django.db.backends.utils.CursorWrapper.

    It exposes the same public interface, but logs executed queries
    to the provided logger before delegating them to the wrapped cursor.
    """

    def __init__(
        self,
        cursor,
        db,
        logger,
        collect_stacktrace=None,
        query_explain: bool = False,  # Capture EXPLAIN SQL information?
    ):
        self.cursor = cursor
        self.db = db
        self.logger = logger  # must implement 'record' method

        if collect_stacktrace is None:
            self.get_stacktrace = get_stacktrace
        else:
            self.get_stacktrace = collect_stacktrace

        self.query_explain = query_explain

    def __getattr__(self, attr):
        return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _quote_expr(self, element):
        if isinstance(element, str):
            return "'%s'" % element.replace("'", "''")
        else:
            return repr(element)

    def _quote_params(self, params):
        if not params:
            return params
        if isinstance(params, dict):
            return {key: self._quote_expr(value) for key, value in params.items()}
        return [self._quote_expr(p) for p in params]

    def _decode(self, param):
        if PostgresJson and isinstance(param, PostgresJson):
            return param.dumps(param.adapted)

        # decode each item separately for data containers
        if isinstance(param, (tuple, list)):
            return [self._decode(element) for element in param]
        if isinstance(param, dict):
            return {key: self._decode(value) for key, value in param.items()}

        # make sure datetime, date and time are converted to string by force_str
        CONVERT_TYPES = (datetime.datetime, datetime.date, datetime.time)
        try:
            return force_str(param, strings_only=not isinstance(param, CONVERT_TYPES))
        except UnicodeDecodeError:
            return repr(param)

    def _record(self, method, sql, params):
        if not self.query_explain:
            explain = None
        else:
            explain_prefix = self.db.ops.explain_query_prefix()
            self.cursor.execute(f'{explain_prefix} {sql}', params)
            result = self.cursor.fetchall()

            # Convert the result in the same way as Django, see: SQLCompiler.explain_query():
            explain = []
            for row in result:
                if not isinstance(row, str):
                    explain.append(' '.join(str(c) for c in row))
                else:
                    explain.append(row)

        start = time.monotonic()
        try:
            return method(sql, params)
        finally:
            stop = time.monotonic()
            duration = (stop - start) * 1000

            try:
                _params_decoded = json.dumps(self._decode(params), cls=DjangoJSONEncoder)
            except TypeError:
                _params_decoded = ''  # object not JSON serializable, we have to live with that

            sql = str(sql)  # is sometimes an object, e.g. psycopg Composed, so ensure string
            stacktrace = self.get_stacktrace()

            self.logger.record(
                alias=getattr(self.db, 'alias', 'default'),
                vendor=getattr(self.db.connection, 'vendor', 'unknown'),
                raw_sql=sql,
                sql=self.db.ops.last_executed_query(self.cursor, sql, self._quote_params(params)),
                raw_params=params,
                params=_params_decoded,
                duration=duration,
                stacktrace=stacktrace,
                explain=explain,
            )

    def callproc(self, procname, params=None):
        return self._record(self.cursor.callproc, procname, params)

    def execute(self, sql, params=None):
        return self._record(self.cursor.execute, sql, params)

    def executemany(self, sql, param_list):
        return self._record(self.cursor.executemany, sql, param_list)



from copy import deepcopy

from django.conf import settings
from django.core.cache import cache
from django.test import override_settings

from bx_django_utils.feature_flags.data_classes import FeatureFlag
from bx_django_utils.feature_flags.models import FeatureFlagModel
from bx_django_utils.test_utils.cache import ClearCacheMixin


def get_feature_flag_states() -> dict:
    """
    Collects information about all registered feature flags and their current state.
    If FeatureFlagTestCaseMixin used -> the current state is the initial state!
    """
    return {feature_flag.cache_key: feature_flag.is_enabled for feature_flag in FeatureFlag.values()}


def get_feature_flag_cache_info() -> dict:
    # Not in README: Probably only useful for internal use.
    result = {}
    for cache_key in sorted(FeatureFlag.registry.keys()):
        state = cache.get(cache_key)
        if state is not None:
            result[cache_key] = state
    return result


def get_feature_flag_db_info() -> dict:
    # Not in README: Probably only useful for internal use.
    return dict(FeatureFlagModel.objects.order_by('cache_key').values_list('cache_key', 'state'))


class FeatureFlagTestCaseMixin(ClearCacheMixin):
    """
    Mixin for `TestCase` that will change `FeatureFlag` entries. To make the tests atomic.

    * Restore `FeatureFlag.registry` between tests.
    * Restore all feature flags to the initial stage between tests.

    Used ClearCacheMixin to remove persistent data in Django's cache.
    Should be used with django.test.TestCase so that persistent data in database also be removed!
    """

    @classmethod
    def setUpClass(cls):
        cls._origin_feature_flag_registry = FeatureFlag.registry
        super().setUpClass()

        # Use per-process and thread-safe cache backend for tests.
        # This is important if tests run with `--parallel`
        caches_settings = deepcopy(settings.CACHES)
        caches_settings['default'] = {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            'LOCATION': 'unique-snowflake',
        }
        cls._overwrite_cache = override_settings(CACHES=caches_settings)
        cls._overwrite_cache.__enter__()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls._overwrite_cache.__exit__(None, None, None)

    def setUp(self):
        super().setUp()
        FeatureFlag.registry = deepcopy(self._origin_feature_flag_registry)

    def tearDown(self):
        super().tearDown()
        FeatureFlag.registry = self._origin_feature_flag_registry



from django.contrib.auth import get_user_model
from django.test import SimpleTestCase, TestCase
from manageprojects.test_utils.project_setup import check_editor_config, get_py_max_line_length
from model_bakery import baker
from packaging.version import Version

from bx_django_utils import __version__
from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin
from bx_django_utils_tests.tests import PACKAGE_ROOT


class ProjectSetupTestCase(SimpleTestCase):
    def test_version(self):
        self.assertIsNotNone(__version__)

        version = Version(__version__)  # Will raise InvalidVersion() if wrong formatted
        self.assertEqual(str(version), __version__)

    def test_check_editor_config(self):
        check_editor_config(package_root=PACKAGE_ROOT)

        max_line_length = get_py_max_line_length(package_root=PACKAGE_ROOT)
        self.assertEqual(max_line_length, 119)


class DjangoProjectSetupTestCase(HtmlAssertionMixin, TestCase):
    def test_admin(self):
        response = self.client.get('/admin/')
        self.assertRedirects(response, expected_url='/admin/login/?next=/admin/')

        UserModel = get_user_model()
        user = baker.make(UserModel, is_staff=True, is_active=True, is_superuser=True)
        self.client.force_login(user)

        with self.settings(DEBUG=True):
            response = self.client.get('/admin/')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, template_name='admin/index.html')
        self.assert_html_parts(
            response,
            parts=(
                '<title>Site administration | Django site admin</title>',
                '<link rel="stylesheet" href="/static/debug_toolbar/css/toolbar.css">',
            ),
        )



# Generated by Django 4.1.5 on 2023-02-03 11:42

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="FeatureFlagModel",
            fields=[
                (
                    "create_dt",
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        help_text="ModelTimetrackingMixin.create_dt.help_text",
                        null=True,
                        verbose_name="ModelTimetrackingMixin.create_dt.verbose_name",
                    ),
                ),
                (
                    "update_dt",
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        help_text="ModelTimetrackingMixin.update_dt.help_text",
                        null=True,
                        verbose_name="ModelTimetrackingMixin.update_dt.verbose_name",
                    ),
                ),
                (
                    "cache_key",
                    models.CharField(
                        help_text="FeatureFlagModel.cache_key.help_text",
                        max_length=250,
                        primary_key=True,
                        serialize=False,
                        verbose_name="FeatureFlagModel.cache_key.verbose_name",
                    ),
                ),
                (
                    "state",
                    models.SmallIntegerField(
                        choices=[(1, "enabled"), (0, "disabled")],
                        help_text="FeatureFlagModel.state.help_text",
                        verbose_name="FeatureFlagModel.state.verbose_name",
                    ),
                ),
            ],
            options={
                "verbose_name": "FeatureFlagModel.verbose_name",
                "verbose_name_plural": "FeatureFlagModel.verbose_name_plural",
            },
        ),
    ]



import dataclasses
from typing import Any
from uuid import UUID, uuid4

import django
from django.core.cache import cache
from django.test import SimpleTestCase

from bx_django_utils.cached_dataclasses import CachedDataclassBase
from bx_django_utils.test_utils.cache import ClearCacheMixin


@dataclasses.dataclass
class ExampleCachedDataclass(CachedDataclassBase):
    integer: int
    dictionary: dict
    uuid: UUID = dataclasses.field(default_factory=uuid4)


@dataclasses.dataclass
class NoUUID(CachedDataclassBase):
    pass


@dataclasses.dataclass
class NoDefaultUUID(CachedDataclassBase):
    uuid: Any = None


class TestCachedDataclass(ClearCacheMixin, SimpleTestCase):
    def test_generate_cache_key(self):
        self.assertEqual(
            ExampleCachedDataclass.generate_cache_key(uuid=UUID(int=1)),
            'ExampleCachedDataclass-00000000-0000-0000-0000-000000000001',
        )

    def test_store_restore_to_cache(self):
        data = ExampleCachedDataclass(integer=1, dictionary={'foo': 'bar'})
        uuid_value = data.uuid
        cache_key = ExampleCachedDataclass.generate_cache_key(uuid=uuid_value)

        # Nothing cached, yet?
        self.assertIsNone(cache.get(cache_key))
        self.assertIsNone(ExampleCachedDataclass.get_from_cache(uuid=uuid_value))

        # Add data to Django's cache:
        data.store2cache()

        # Cached filled?
        cached_data = cache.get(cache_key)
        self.assertEqual(
            cached_data,
            {'dictionary': {'foo': 'bar'}, 'integer': 1, 'uuid': uuid_value},
        )

        # Get data back from cache:
        data2 = ExampleCachedDataclass.get_from_cache(uuid=uuid_value)
        self.assertIsInstance(data2, ExampleCachedDataclass)
        self.assertEqual(data, data2)

        # Delete cache entry:
        deleted = data.delete_cache_entry()
        if django.VERSION >= (3, 1):
            # https://docs.djangoproject.com/en/3.2/topics/cache/#django.core.caches.cache.delete
            self.assertTrue(deleted)
        self.assertIsNone(ExampleCachedDataclass.get_from_cache(uuid=uuid_value))

    def test_no_uuid_argument(self):
        with self.assertRaisesMessage(AssertionError, 'Child class must define a "uuid" value!'):
            NoUUID()

        with self.assertRaisesMessage(AssertionError, 'A UUID is needed. Found: NoneType'):
            NoDefaultUUID(uuid=None)

        uuid_value = UUID(int=0)
        data = NoDefaultUUID(uuid=uuid_value)
        self.assertIsInstance(data, NoDefaultUUID)
        self.assertEqual(data.uuid, UUID('00000000-0000-0000-0000-000000000000'))



import datetime
from unittest import mock

from django.test import TestCase
from django.utils import timezone

from bx_django_utils_tests.test_app.models import TimetrackingTestModel


class TimetrackingMixinTestCase(TestCase):
    def test_timetracking(self):
        item = TimetrackingTestModel()

        mocked_dt1 = datetime.datetime(2000, 1, 2, 3, 4, 5, tzinfo=datetime.timezone.utc)
        with mock.patch.object(timezone, 'now', return_value=mocked_dt1):
            item.save()
        item = TimetrackingTestModel.objects.get(pk=item.pk)
        assert item.create_dt == mocked_dt1
        assert item.update_dt == mocked_dt1

        mocked_dt2 = datetime.datetime(2000, 1, 2, 4, 4, 5, tzinfo=datetime.timezone.utc)
        with mock.patch.object(timezone, 'now', return_value=mocked_dt2):
            item.save()
        item = TimetrackingTestModel.objects.get(pk=item.pk)
        assert item.create_dt == mocked_dt1
        assert item.update_dt == mocked_dt2

        mocked_dt3 = datetime.datetime(2000, 1, 2, 5, 4, 5, tzinfo=datetime.timezone.utc)
        with mock.patch.object(timezone, 'now', return_value=mocked_dt3):
            item.save(update_dt=False)
            item = TimetrackingTestModel.objects.get(pk=item.pk)
            assert item.create_dt == mocked_dt1
            assert item.update_dt == mocked_dt2
            item.save(update_dt=True, update_fields=())
        item = TimetrackingTestModel.objects.get(pk=item.pk)
        assert item.create_dt == mocked_dt1
        assert item.update_dt == mocked_dt3



from unittest import mock

from bx_py_utils.test_utils.mock_uuid import MockUUIDGenerator
from django.test import TestCase

from bx_django_utils.test_utils.users import assert_user_properties, filter_permission_names, make_test_user


class UserUtilsTestCase(TestCase):
    def test_assert_user_properties_wrong_user_object(self):
        class Wrong:
            pass

        with self.assertRaisesMessage(
            AssertionError, "Given user object (type='Wrong') is not a 'auth.User' instance."
        ):
            assert_user_properties(user=Wrong(), properties={})

    def test_make_test_user1(self):
        with mock.patch('uuid.uuid4', MockUUIDGenerator()):
            user = make_test_user()

        # we can check only a few properties:
        assert_user_properties(
            user, properties={'username': 'test 89e6b14d-622a-409f-88de-000000000001'}
        )
        assert_user_properties(user, properties={'email': 'test@test.tld', 'permissions': []})

        # Check everything:
        assert_user_properties(
            user,
            properties={
                'username': 'test 89e6b14d-622a-409f-88de-000000000001',
                'email': 'test@test.tld',
                'is_active': True,
                'is_staff': True,
                'is_superuser': False,
                'permissions': [],
            },
            raw_password='t',
        )

    def test_make_test_user_permissions(self):
        user = make_test_user(
            permissions=filter_permission_names(['auth.add_user', 'auth.change_user']),
        )
        assert_user_properties(
            user,
            properties={
                'permissions': ['auth.add_user', 'auth.change_user'],
            },
        )

    def test_make_test_user_flags(self):
        assert_user_properties(
            user=make_test_user(is_active=False),
            properties={'is_active': False},
        )
        assert_user_properties(
            user=make_test_user(is_staff=False),
            properties={'is_staff': False},
        )
        assert_user_properties(
            user=make_test_user(is_superuser=True),
            properties={'is_superuser': True},
        )



import datetime

from bx_py_utils.test_utils.datetime import parse_dt
from django.test import SimpleTestCase
from django.utils import translation

from bx_django_utils.templatetags.humanize_time import human_duration


class HumanizeTimeTestCase(SimpleTestCase):
    def test_basic(self):
        with translation.override('en'):
            result = human_duration(
                parse_dt('2000-01-01T12:00:00+0000'),
                parse_dt('2000-01-01T12:10:00+0000'),
            )
            assert result == '<span title="Jan. 1, 2000, noon">10.0\xa0minutes</span>'

        with translation.override('en'):
            result = human_duration(
                parse_dt('2000-01-01T12:00:00+0000'),
                parse_dt('2000-01-01T12:00:02+0000'),
            )
            assert result == '<span title="Jan. 1, 2000, noon">2.0\xa0seconds</span>'

        with translation.override('de'):
            result = human_duration(
                parse_dt('2000-01-01T12:00:00+0000'),
                parse_dt('2000-01-01T12:10:00+0000'),
            )
            self.assertEqual(result, '<span title="1. Januar 2000 12:00">10.0\xa0Minuten</span>')

        with translation.override('en'):
            result = human_duration(
                parse_dt('2000-01-01T12:32:12+0000'),
                parse_dt('2000-01-01T12:20:10+0000'),
            )
            assert result == '<span title="Jan. 1, 2000, 12:32 p.m.">-12.0\xa0minutes</span>'

        with translation.override('en'):
            years_back = datetime.datetime.now() - datetime.timedelta(days=5 * 365)
            result = human_duration(years_back)
            assert result.endswith('>5.0\xa0years</span>')

        assert human_duration(None) == ''
        assert human_duration(value=object) == ''

        with translation.override('en'):
            result = human_duration(
                datetime.date(2000, 1, 1),
                datetime.date(2000, 6, 15),
            )
            assert result == '<span title="Jan. 1, 2000, midnight">5.5\xa0months</span>'



import logging

from django.db import models

from bx_django_utils.approve_workflow.models import BaseApproveModel, BaseApproveWorkflowModel


logger = logging.getLogger(__name__)


class ApproveTestModel(BaseApproveWorkflowModel):
    """
    Example implementation of a ApproveWorkflowModel with a relation.
    Used in tests.
    """
    name = models.CharField(
        max_length=128,
        verbose_name='The Name',
        help_text='This is always needed field',
    )
    title = models.CharField(
        max_length=128,
        blank=True,
        verbose_name='A Title',
        help_text='This is only needed to approve this entry'
    )

    # Fields are required for *draft* instances (used in admin model form):
    REQUIRED_FIELDS_DRAFT = ()

    # Fields are required for *approve* instances (used in admin model form):
    REQUIRED_FIELDS_PUBLIC = ('title',)

    def _copy_normal_fields(self, approve_instance):
        """
        Just assign all normal field values from draft to approve version.
        """
        approve_instance.name = self.name
        approve_instance.title = self.title

    def _copy_relations(self, approve_instance):
        """
        Assign relations from draft to approve version.
        """
        draft_relations = self.relations.all()
        seen_pks = []
        for draft_relation in draft_relations:
            seen_pks.append(draft_relation.pk)
            approved_relation = draft_relation.approve(main_entry_id=approve_instance.pk)
            seen_pks.append(approved_relation.pk)

        # Remove obsolete relations:
        cleanup_qs = RelatedApproveTestModel.objects.filter(
            main_entry=approve_instance
        ).exclude(pk__in=seen_pks)
        delete_count = cleanup_qs.delete()[0]
        if delete_count:
            logger.debug('delete %i relations', delete_count)


class RelatedApproveTestModel(BaseApproveModel):
    relation_name = models.CharField(
        max_length=128,
        verbose_name='The Name',
        help_text='This is always needed field',
    )
    relation_title = models.CharField(
        max_length=128,
        null=True,
        blank=True,
        verbose_name='A Title',
        help_text='This is only needed to approve this entry'
    )

    main_entry = models.ForeignKey(
        ApproveTestModel,
        on_delete=models.CASCADE,
        related_name='relations',
    )

    # Fields are required for *draft* instances (used in admin model form):
    REQUIRED_FIELDS_DRAFT = ()

    # Fields are required for *approve* instances (used in admin model form):
    REQUIRED_FIELDS_PUBLIC = ('relation_title',)

    def _copy_normal_fields(self, approve_instance):
        """
        Just assign all normal field values from draft to approve version.
        """
        approve_instance.relation_name = self.relation_name
        approve_instance.relation_title = self.relation_title

    def _copy_relations(self, approve_instance):
        """
        Assign relations from draft to approve version.
        """
        approve_instance.main_entry = self.main_entry

    def approve(self, main_entry_id):
        approve_instance = super().approve(
            # Publish and attach to given main_entry:
            main_entry_id=main_entry_id
        )

        return approve_instance



"""
    Utilities to manipulate objects in database via models:
    e.g.:
        create/update/delete model entries etc.
"""
import dataclasses
import warnings
from typing import Any, Callable, Optional

from django.core.exceptions import FieldDoesNotExist
from django.db import models


STORE_BEHAVIOR_IGNORE = "i"  # Ignore the value completely
STORE_BEHAVIOR_SET_IF_EMPTY = "e"  # Use the value only if we currently have no one.
STORE_BEHAVIOR_SKIP_EMPTY = "s"  # Don't store empty values (and maybe overwrite existing one)


class InvalidStoreBehavior(FieldDoesNotExist):
    """
    Exception used in create_or_update() if "store_behavior" contains not existing field names.
    """
    pass


def create(*, ModelClass, call_full_clean=True, save_kwargs=None, **values):
    """
    Create a new model instance with optional validate before create.
    """
    instance = ModelClass(**values)
    if call_full_clean:
        instance.full_clean(validate_unique=False)  # Don't create non-valid instances
    if save_kwargs is None:
        save_kwargs = {}
    instance.save(force_insert=True, **save_kwargs)
    return instance


@dataclasses.dataclass
class FieldUpdate:
    """
    Information about updated model field values. Used for CreateOrUpdateResult.update_info
    """

    field_name: str
    old_value: Any
    new_value: Any


@dataclasses.dataclass
class CreateOrUpdateResult:
    """
    Result object returned by create_or_update2() with all information about create/save a model.
    Contains:
        Model instance, that is created or updated.
        List of model field names that are: Updated / ignored / not overwritten
    """
    # Model instance object that was created or updated:
    instance: type[models.Model] = None

    # If True: new instance was created else: existing was updated:
    created: bool = False

    # Fields that are updated (Empty list if model was created!):
    updated_fields: list = dataclasses.field(default_factory=list)

    # Old and new values if a existing instance was updates (Empty if new instance created):
    update_info: list[FieldUpdate] = dataclasses.field(default_factory=list)

    # Field names that ignored by "STORE_BEHAVIOR_IGNORE":
    ignored_fields: list = dataclasses.field(default_factory=list)

    # Field names that are not overwritten (STORE_BEHAVIOR_SET_IF_EMPTY):
    not_overwritten_fields: list = dataclasses.field(default_factory=list)

    # Field names that are not filles with empty value (STORE_BEHAVIOR_SKIP_EMPTY):
    skip_empty_values: list = dataclasses.field(default_factory=list)


def update_model_field(*, instance, field_name, old_value, new_value, result: CreateOrUpdateResult):
    """
    Default callback for create_or_update2() to set a changed model field value and expand CreateOrUpdateResult
    """
    if old_value == new_value:
        # Nothing to update -> we are done.
        return

    # The model field value has been changed -> set it on the model instance:
    setattr(instance, field_name, new_value)

    # Expand the list of updated fields, used for result and save() call:
    result.updated_fields.append(field_name)

    # Store update information:
    result.update_info.append(FieldUpdate(field_name=field_name, old_value=old_value, new_value=new_value))


def create_or_update2(
    *,
    ModelClass: type[models.Model],
    lookup: dict = None,
    call_full_clean: bool = True,
    store_behavior: Optional[dict] = None,
    save_kwargs: Optional[dict] = None,
    update_model_field_callback: Callable = update_model_field,
    **values,
) -> CreateOrUpdateResult:
    """
    Create a new model instance or update a existing one and returns CreateOrUpdateResult instance
    Similar to django's own create_or_update() but:

     * Use "update_fields" in save() call to store only changed fields.
     * return the information about changed field names.
     * validate before save.
     * Accept optional "store_behavior"
     * Accept optional callback for every updated model field value

     "store_behavior" is a dict with information about overwriting field values, e.g.:

         store_behavior = {
            'field_name1': STORE_BEHAVIOR_IGNORE,
            'field_name2': STORE_BEHAVIOR_SET_IF_EMPTY,
            'field_name3': STORE_BEHAVIOR_SKIP_EMPTY,
            ...
         }

     behaviors are:
      - STORE_BEHAVIOR_IGNORE........: Never store a given value to the model field
      - STORE_BEHAVIOR_SET_IF_EMPTY..: Store given value only if current field value is empty
      - STORE_BEHAVIOR_SKIP_EMPTY....: Don't store empty values to the field (protect existing)
    """
    if save_kwargs is None:
        save_kwargs = {}
    result = CreateOrUpdateResult()

    to_be_ignore_fields = set()  # Fields that should be ignored.
    to_be_set_if_empty_fields = set()  # Fields that should are set only when empty.
    to_be_skiped_if_empty = set()  # Fields that should not filled with empty values

    if store_behavior:
        opts = ModelClass._meta
        all_field_names = {
            # Note: We collect intentionally all fields + relations etc.
            # Maybe relations will be handled in external code parts ;)
            field.name
            for field in opts.get_fields(include_parents=True, include_hidden=True)
        }
        for field_name, behavior in store_behavior.items():
            if field_name not in all_field_names:
                raise InvalidStoreBehavior(
                    f'store_behavior field name {field_name!r}'
                    f' is not one of: {sorted(all_field_names)}'
                )

            if behavior == STORE_BEHAVIOR_IGNORE:
                # Values for this field should be completely ignored
                to_be_ignore_fields.add(field_name)
            elif behavior == STORE_BEHAVIOR_SET_IF_EMPTY:
                # Field values should be only stored, if existing field is empty
                to_be_set_if_empty_fields.add(field_name)
            elif behavior == STORE_BEHAVIOR_SKIP_EMPTY:
                # Fields that should not filled with empty values
                to_be_skiped_if_empty.add(field_name)
            else:
                raise KeyError(f'Unknown store behavior: {behavior!r} !')

    if not to_be_ignore_fields:
        filtered_values = values
    else:
        # Just filter every field values that we should ignore:
        filtered_values = {}
        for key, value in values.items():
            if key in to_be_ignore_fields:
                result.ignored_fields.append(key)
            else:
                filtered_values[key] = value

    if lookup is None:
        # Create a new object
        instance = create(
            ModelClass=ModelClass, call_full_clean=call_full_clean, save_kwargs=save_kwargs,
            **filtered_values)
        result.instance = instance
        result.created = True
        return result

    # Try to update a existing object
    assert isinstance(lookup, dict)
    instance = ModelClass.objects.filter(**lookup).first()
    if not instance:
        instance = create(
            ModelClass=ModelClass,
            call_full_clean=call_full_clean,
            save_kwargs=save_kwargs,
            **lookup,
            **filtered_values
        )
        result.instance = instance
        result.created = True
        return result

    # Store values:

    for field_name, value in filtered_values.items():
        if not value and field_name in to_be_skiped_if_empty:
            # We should not store empty value (and maybe overwrite existing one):
            result.skip_empty_values.append(field_name)
            continue

        old_value = getattr(instance, field_name)

        if old_value and field_name in to_be_set_if_empty_fields:
            # We should not overwrite this existing field value!
            result.not_overwritten_fields.append(field_name)
            continue

        update_model_field_callback(
            instance=instance,
            field_name=field_name,
            old_value=old_value,
            new_value=value,
            result=result,
        )

    if result.updated_fields:
        if call_full_clean:
            instance.full_clean(validate_unique=False)  # Don't save new non-valid values
        instance.save(update_fields=result.updated_fields, **save_kwargs)

    result.instance = instance

    return result


def create_or_update(*, ModelClass: type[models.Model], lookup: dict = None, call_full_clean: bool = True, **values):
    """
    Create a new model instance or update a existing one. Deprecated! Use: create_or_update2()
    """
    warnings.warn(
        'create_or_update() is deprecated, please use create_or_update2()',
        DeprecationWarning
    )
    result = create_or_update2(
        ModelClass=ModelClass, lookup=lookup, call_full_clean=call_full_clean, **values
    )
    instance = result.instance
    created = result.created
    updated_fields = result.updated_fields or None  # mimic old return value: None and not []
    return instance, created, updated_fields



from django.test import TestCase
from django.urls import reverse

from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin
from bx_django_utils.test_utils.users import make_minimal_test_user


class DynamicViewMenuTestCase(HtmlAssertionMixin, TestCase):
    def test_menu(self):
        minimal_test_user = make_minimal_test_user(
            is_staff=False,
            permissions=()
        )
        self.client.force_login(user=minimal_test_user)

        # User sees the test app model:
        response = self.client.get(path='/')
        self.assert_html_parts(response, parts=(
            '<title>Dynamic View Menu - Index | Django site admin</title>',
            '''
            <div id="content" class="colM">
              <h1>Dynamic View Menu - Index</h1>
            <h2>Section 1</h2>
            <ul>
                <li><a href="/redirect2admin/">Go to Django Admin</a></li>
            </ul>
            <h2>Section 2</h2>
            <ul>
                <li><a href="/demo1/">Dynamic View Menu - DEMO 1</a></li>
                <li><a href="/demo2/">Dynamic View Menu - DEMO 2</a></li>
            </ul>
            <h2>Section 3</h2>
            <ul>
                <li><a href="/demo3/">Dynamic View Menu - DEMO 3</a></li>
            </ul>
              <br class="clear">
            </div>
            '''
        ))

        response = self.client.get(path='/redirect2admin/')
        self.assertRedirects(
            response,
            expected_url='/admin/',
            target_status_code=302  # redirect to login, because we are not a staff user ;)
        )

        response = self.client.get(path='/demo2/')
        self.assert_html_parts(response, parts=(
            '<title>Dynamic View Menu - DEMO 2 | Django site admin</title>',
            '''
            <div id="content" class="colM">
                <h1>Dynamic View Menu - DEMO 2</h1>
                DEMO 2 content
                <br class="clear">
            </div>
            '''
        ))

        # registered with namespace?
        assert reverse('test_app:redirect2admin') == '/redirect2admin/'
        assert reverse('test_app:demo1') == '/demo1/'
